from enum import Enum

class Vistock24HMoneyIndustryCategory(Enum):
    OIL_AND_GAS = '0500'
    CHEMICALS = '1300'
    BASIC_RESOURCES = '1700'
    CONSTRUCTION_AND_MATERIALS = '2300'
    INDUSTRIAL_GOODS_AND_SERVICES = '2700'
    AUTOMOBILES_AND_PARTS = '3300'
    FOOD_AND_BEVERAGE = '3500'
    PERSONAL_HOUSEHOLD_GOODS = '3700'
    HEALTH_CARE = '4500'
    RETAIL = '5300'
    MEDIA = '5500'
    TRAVEL_AND_LEISURE = '5700'
    TELECOMMUNICATIONS = '6500'
    ULTILITIES = '7500' # Electrticity, Water, Fuel & Gas
    BANKS = '8300'
    INSURANCE = '8500'
    REAL_ESTATE = '8600'
    FINANCIAL_SERVICES = '8700'
    INFORMATION_TECHNOLOGY = '9500'
    ALL = 'all'

class Vistock24HMoneyFloorCategory(Enum):
    HOSE = 'HOSE'
    HNX = 'HNX'
    UPCOM = 'UPCOM'
    ALL = 'all'

class Vistock24HMoneyCompanyCategory(Enum):
    COMMON_TECHNOLOGY_SHARES = 'CT'
    BANKING = 'NH'
    INSURANCE = 'BH'
    SECURITIES = 'CK'
    INVESTMENT_FUND = 'FUND'
    ALL = 'all'

class Vistock24HMoneyLetterCategory(Enum):
    A = 'A'
    B = 'B'
    C = 'C'
    D = 'D'
    E = 'E'
    F = 'F'
    G = 'G'
    H = 'H'
    I = 'I'
    J = 'J'
    K = 'K'
    L = 'L'
    M = 'M'
    N = 'N'
    O = 'O'
    P = 'P'
    Q = 'Q'
    R = 'R'
    S = 'S'
    T = 'T'
    U = 'U'
    V = 'V'
    W = 'W'
    X = 'X'
    Y = 'Y'
    Z = 'Z'
    ALL = 'all'

class Vistock24HMoneySectionMapping(Enum):
    APPAREL_RETAIL = 'Bán lẻ hàng may mặc'
    COMPLEX_RETAIL = 'Bán lẻ phức hợp'
    LIFE_INSURANCE = 'Bảo hiểm nhân thọ'
    NON_LIFE_INSURANCE = 'Bảo hiểm phi nhân thọ'
    REAL_ESTATE = 'Bất động sản'
    EXPRESS_DELIVERY = 'Chuyển phát nhanh'
    HEALTHCARE = 'Chăm sóc y tế'
    WASTE_ENVIRONMENT = 'Chất thải & Môi trường'
    CONTAINERS_PACKAGING = 'Containers & Đóng gói'
    COMPLEX_INDUSTRIES = 'Công nghiệp phức hợp'
    BIOTECHNOLOGY = 'Công nghệ sinh học'
    PHARMACEUTICALS = 'Dược phẩm'
    COMPUTER_SERVICES = 'Dịch vụ Máy tính'
    ENTERTAINMENT_SERVICES = 'Dịch vụ giải trí'
    SPECIALTY_CONSUMER_SERVICES = 'Dịch vụ tiêu dùng chuyên ngành'
    MEDIA_SERVICES = 'Dịch vụ truyền thông'
    TRANSPORTATION_SERVICES = 'Dịch vụ vận tải'
    MEDICAL_INSTRUMENTS = 'Dụng cụ y tế'
    ENTERTAINMENT_MEDIA = 'Giải trí & Truyền thông'
    FOOTWEAR = 'Giầy dép'
    APPAREL = 'Hàng May mặc'
    PERSONAL_GOODS = 'Hàng cá nhân'
    AVIATION = 'Hàng không'
    ELECTRICAL_ELECTRONICS = 'Hàng điện & điện tử'
    INTERNET = 'Internet'
    MINING = 'Khai khoáng '
    COAL_MINING = 'Khai thác Than'
    GOLD_MINING = 'Khai thác vàng'
    LOGISTICS = 'Kho bãi, hậu cần và bảo dưỡng'
    HOTELS = 'Khách sạn'
    NON_FERROUS_METALS = 'Kim Loại màu'
    TIMBER_WOOD = 'Lâm sản và Chế biến gỗ'
    TIRES = 'Lốp xe'
    INDUSTRIAL_MACHINERY = 'Máy công nghiệp'
    SECURITIES_BROKERAGE = 'Môi giới chứng khoán'
    BANKING = 'Ngân hàng'
    EQUIPMENT_SUPPLIERS = 'Nhà cung cấp thiết bị'
    RESTAURANTS_BARS = 'Nhà hàng và quán bar'
    ALUMINUM = 'Nhôm'
    PLASTICS_RUBBER_FIBERS = 'Nhựa, cao su & sợi'
    AGRICULTURE_AQUACULTURE = 'Nuôi trồng nông & hải sản'
    WATER = 'Nước'
    PHARMA_DISTRIBUTION = 'Phân phối dược phẩm'
    SPECIALTY_DISTRIBUTION = 'Phân phối hàng chuyên dụng'
    FOOD_DISTRIBUTION = 'Phân phối thực phẩm'
    FUEL_GAS_DISTRIBUTION = 'Phân phối xăng dầu & khí đốt'
    HARDWARE = 'Phần cứng'
    SOFTWARE = 'Phần mềm'
    AUTO_PARTS = 'Phụ tùng ô tô'
    ASSET_MANAGEMENT = 'Quản lý tài sản'
    INVESTMENT_FUNDS = 'Quỹ đầu tư'
    BOOKS_CULTURE = 'Sách, ấn bản & sản phẩm văn hóa'
    CHEMICALS = 'Sản phẩm hóa dầu, Nông dược & Hóa chất khác'
    ELECTRICITY = 'Sản xuất & Phân phối Điện'
    BEER_PRODUCTION = 'Sản xuất bia '
    PAPER_PRODUCTION = 'Sản xuất giấy'
    OIL_GAS = 'Sản xuất và Khai thác dầu khí'
    AUTOMOBILE_MANUFACTURING = 'Sản xuất ô tô'
    HOME_APPLIANCES = 'Thiết bị gia dụng'
    TELECOM_EQUIPMENT = 'Thiết bị viễn thông'
    OIL_GAS_EQUIP_SERVICES = 'Thiết bị và Dịch vụ Dầu khí'
    OFFICE_EQUIPMENT = 'Thiết bị văn phòng'
    MEDICAL_EQUIPMENT = 'Thiết bị y tế'
    ELECTRICAL_EQUIPMENT = 'Thiết bị điện'
    TOBACCO = 'Thuốc lá'
    STEEL_PRODUCTS = 'Thép và sản phẩm thép'
    FOOD = 'Thực phẩm'
    OTHER_UTILITIES = 'Tiện ích khác'
    PERSONAL_FINANCE = 'Tài chính cá nhân'
    SPECIAL_FINANCE = 'Tài chính đặc biệt'
    REINSURANCE = 'Tái bảo hiểm'
    REAL_ESTATE_CONSULTING = 'Tư Vấn, Định giá, Môi giới Bất động sản'
    BUSINESS_CONSULTING = 'Tư vấn & Hỗ trợ KD'
    WINE_SPIRITS = 'Vang & Rượu mạnh'
    FIXED_TELECOM = 'Viễn thông cố định'
    MOBILE_TELECOM = 'Viễn thông di động'
    WATER_TRANSPORT = 'Vận tải Thủy'
    PASSENGER_TRAVEL = 'Vận tải hành khách & Du lịch'
    BUILDING_MATERIALS = 'Vật liệu xây dựng & Nội thất'
    TRUCKS_SHIPBUILDING = 'Xe tải & Đóng tàu'
    CONSTRUCTION = 'Xây dựng'
    CONSUMER_ELECTRONICS = 'Điện tử tiêu dùng'
    TRAINING_EMPLOYMENT = 'Đào tạo & Việc làm'
    TOYS = 'Đồ chơi'
    DURABLE_HOUSEHOLD_GOODS = 'Đồ gia dụng lâu bền'
    DISPOSABLE_HOUSEHOLD_GOODS = 'Đồ gia dụng một lần'
    BEVERAGES_REFRESHMENTS = 'Đồ uống & giải khát'

class VistockVnDirectFinancialModelsCategory(Enum):
    BALANCE_SHEET = '1,89,101,411'
    INCOME_STATEMENT = '2,90,102,412'
    CASH_FLOW_STATEMENT = '3,91,103,413'
    ALL = 'all'

class VistockVnDirectReportTypeCategory(Enum):
    ANNUAL = 'ANNUAL'
    QUARTER = 'QUARTER'

class VistockVnDirectIndexCodeMapping(Enum):
    VNINDEX = 'VNINDEX'
    HNX = 'HNX'
    UPCOM = 'UPCOM'
    VN30 = 'VN30'
    VN30F1M = 'VN30F1M'

class VistockVnDirectChangePricePeriodMapping(Enum):
    ONE_DAY = '1D'
    MONTH_TO_DATE = 'MTD'
    QUARTER_TO_DATE = 'QTD'
    YEAR_TO_DATE = 'YTD'
    ONE_WEEK = '5D'
    ONE_MONTH = '1M'
    THREE_MONTHS = '3M'
    SIX_MONTHS = '6M'
    ONE_YEAR = '1Y'


from vistock.core.models import (
    StandardVnDirectStockIndexSearchResults, 
    AdvancedVnDirectStockIndexSearchResults,
    StandardVnDirectFundamentalIndexSearchResults,
    Standard24HMoneyStockSectionSearchResults,
    StandardVnDirectFinancialModelSearchResults,
    StandardVnDirectFinancialStatementsIndexSearchResults,
    StandardDNSEStockIndexSearchResults,
    StandardVietstockStockIndexSearchResults,
    StandardVnDirectMarketPricesSearchResults,
    StandardVnDirectChangePricesSearchResults
)
from vistock.core.enums import (
    Vistock24HMoneyIndustryCategory,
    Vistock24HMoneyFloorCategory,
    Vistock24HMoneyCompanyCategory,
    Vistock24HMoneyLetterCategory,
    VistockVnDirectFinancialModelsCategory,
    VistockVnDirectReportTypeCategory,
    VistockVnDirectIndexCodeMapping,
    VistockVnDirectChangePricePeriodMapping
)
from typing import Union, Protocol, Literal
from datetime import datetime, timezone

class IVistockVnDirectStockIndexSearch(Protocol):
    def search(
        self, 
        code: str,
        start_date: str = '2012-01-01',
        end_date: str = datetime.now().strftime('%Y-%m-%d'),
        resolution: Literal['day'] = 'day',
        advanced: bool = True,
        ascending: bool = False
    ) -> Union[StandardVnDirectStockIndexSearchResults, AdvancedVnDirectStockIndexSearchResults]:
        ...

class AsyncIVistockVnDirectStockIndexSearch(Protocol):
    async def async_search(
        self, 
        code: str,
        start_date: str = '2012-01-01',
        end_date: str = datetime.now().strftime('%Y-%m-%d'),
        resolution: Literal['day'] = 'day',
        advanced: bool = True,
        ascending: bool = False
    ) -> Union[StandardVnDirectStockIndexSearchResults, AdvancedVnDirectStockIndexSearchResults]:
        ...

class IVistockVnDirectFundamentalIndexSearch(Protocol):
    def search(
        self,
        code: str
    ) -> StandardVnDirectFundamentalIndexSearchResults:
        ...

class AsyncIVistockVnDirectFundamentalIndexSearch(Protocol):
    async def async_search(
        self,
        code: str
    ) -> StandardVnDirectFundamentalIndexSearchResults:
        ...

class IVistockVnDirectFinancialModelsSearch(Protocol):
    def search(
        self,
        code: str,
        model_type: Union[VistockVnDirectFinancialModelsCategory, str] = 'all'
    ) -> StandardVnDirectFinancialModelSearchResults:
        ...

class AsyncIVistockVnDirectFinancialModelsSearch(Protocol):
    async def async_search(
        self,
        code: str,
        model_type: Union[VistockVnDirectFinancialModelsCategory, str] = 'all'
    ) -> StandardVnDirectFinancialModelSearchResults:
        ...

class IVistockVnDirectFinancialStatementsIndexSearch(Protocol):
    def search(
        self,
        code: str,
        start_year: int = 2000,
        end_year: int = datetime.now().year,
        report_type: Union[VistockVnDirectReportTypeCategory, str] = 'ANNUAL',
        model_type: Union[VistockVnDirectFinancialModelsCategory, str] = 'all'
    ) -> StandardVnDirectFinancialStatementsIndexSearchResults:
        ...

class AsyncIVistockVnDirectFinancialStatementsIndexSearch(Protocol):
    async def async_search(
        self,
        code: str,
        start_year: int = 2000,
        end_year: int = datetime.now().year,
        report_type: Union[VistockVnDirectReportTypeCategory, str] = 'ANNUAL',
        model_type: Union[VistockVnDirectFinancialModelsCategory, str] = 'all'
    ) -> StandardVnDirectFinancialStatementsIndexSearchResults:
        ...

class IVistockVnDirectMarketPricesSearch(Protocol):
    def search(
        self,
        code: Union[VistockVnDirectIndexCodeMapping, str],
        start_date: str = '2012-01-01',
        ascending: bool = True
    ) -> StandardVnDirectMarketPricesSearchResults:
        ...

class AsyncIVistockVnDirectMarketPricesSearch(Protocol):
    async def async_search(
        self,
        code: Union[VistockVnDirectIndexCodeMapping, str],
        start_date: str = '2012-01-01',
        ascending: bool = True
    ) -> StandardVnDirectMarketPricesSearchResults:
        ...

class IVistockVnDirectChangePricesSearch(Protocol):
    def search(
        self,
        code: Union[VistockVnDirectIndexCodeMapping, str] = 'VNINDEX,HNX,UPCOM,VN30,VN30F1M',
        period: Union[VistockVnDirectChangePricePeriodMapping, str] = '1D'
    ) -> StandardVnDirectChangePricesSearchResults:
        ...

class AsyncIVistockVnDirectChangePricesSearch(Protocol):
    async def async_search(
        self,
        code: Union[VistockVnDirectIndexCodeMapping, str] = 'VNINDEX,HNX,UPCOM,VN30,VN30F1M',
        period: Union[VistockVnDirectChangePricePeriodMapping, str] = '1D'
    ) -> StandardVnDirectChangePricesSearchResults:
        ...

class IVistock24HMoneyStockSectionSearch(Protocol):
    def search(
        self,
        industry: Union[Vistock24HMoneyIndustryCategory, str] = 'all',
        floor: Union[Vistock24HMoneyFloorCategory, str] = 'all',
        company_type: Union[Vistock24HMoneyCompanyCategory, str] = 'all',
        letter: Union[Vistock24HMoneyLetterCategory, str] = 'all',
        limit: int = 2000
    ) -> Standard24HMoneyStockSectionSearchResults:
        ...

class AsyncIVistock24HMoneyStockSectionSearch(Protocol):
    async def async_search(
        self,
        industry: Union[Vistock24HMoneyIndustryCategory, str] = 'all',
        floor: Union[Vistock24HMoneyFloorCategory, str] = 'all',
        company_type: Union[Vistock24HMoneyCompanyCategory, str] = 'all',
        letter: Union[Vistock24HMoneyLetterCategory, str] = 'all',
        limit: int = 2000 
    ) -> Standard24HMoneyStockSectionSearchResults:
        ...

class IVistockVietstockStockIndexSearch(Protocol):
    def search(
        self,
        code: str,
        resolution: Literal['1D'] = '1D',
        start_date: str = '2000-01-01',
        end_date: str = datetime.now().strftime('%Y-%m-%d'),
        ascending: bool = False
    ) -> StandardVietstockStockIndexSearchResults:
        ...

class AsyncIVistockVietstockStockIndexSearch(Protocol):
    async def async_search(
        self,
        code: str,
        resolution: Literal['1D'] = '1D',
        start_date: str = '2000-01-01',
        end_date: str = datetime.now().strftime('%Y-%m-%d'),
        ascending: bool = False
    ) -> StandardVietstockStockIndexSearchResults:
        ...

class IVistockDNSEStockIndexSearch(Protocol):
    def search(
        self,
        code: str,
        current_datetime: str = datetime.now(timezone.utc).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z',
        ascending: bool = False
    ) -> StandardDNSEStockIndexSearchResults:
        ...

class AsyncIVistockDNSEStockIndexSearch(Protocol):
    async def async_search(
        self,
        code: str,
        current_datetime: str = datetime.now(timezone.utc).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z',
        ascending: bool = False
    ) -> StandardDNSEStockIndexSearchResults:
        ...


from vistock.core.enums import Vistock24HMoneySectionMapping
from typing import List, Dict, Union, Any, Type
from urllib.parse import urlparse
from datetime import datetime
from tzlocal import get_localzone
from enum import Enum
import pytz
import time

class VistockValidator:
    VNDIRECT_STOCK_INDEX_REQUIRED_FIELDS = {
        'code', 'date', 'time', 'floor', 'type', 'basicPrice', 'ceilingPrice',
        'floorPrice', 'open', 'high', 'low', 'close', 'average', 'adOpen',
        'adHigh', 'adLow', 'adClose', 'adAverage', 'nmVolume', 'nmValue',
        'ptVolume', 'ptValue', 'change', 'adChange', 'pctChange'
    }
    DNSE_STOCK_INDEX_REQUIRED_FIELDS = {
        'symbol', 'matchPrice', 'matchQtty', 'sendingTime', 'side'
    }
    VNDIRECT_INDEX_CODES = {
        'VNINDEX', 'HNX', 'UPCOM', 'VN30', 'VN30F1M'
    }
    VNDIRECT_INDEX_PERIODS = {
        '1D', 'MTD', 'QTD', 'YTD', '5D', '1M', '3M', '6M', '1Y'
    }

    @staticmethod
    def validate_url(url: str) -> bool:
        try:
            parsed = urlparse(url)
            return all([parsed.scheme, parsed.netloc])
        
        except Exception:
            return False

    @staticmethod
    def validate_url_with_domain(url: str, domain: str) -> bool:
        try:
            parsed = urlparse(url)
            if parsed.scheme in ('http', 'https') and parsed.hostname and parsed.hostname.endswith(domain):
                return True
            else:
                return False
            
        except Exception:
            return False    
        
    @staticmethod
    def validate_date_format(date_str: str, date_format: str = '%Y-%m-%d') -> bool:        
        try:
            datetime.strptime(date_str, date_format)
            return True
        
        except ValueError:
            return False
        
    @staticmethod
    def validate_date_range(start_date: str, end_date: str, date_format: str = '%Y-%m-%d') -> bool:
        if not (VistockValidator.validate_date_format(start_date, date_format) and 
                VistockValidator.validate_date_format(end_date, date_format)):
            return False
        
        start_dt = datetime.strptime(start_date, date_format)
        end_dt = datetime.strptime(end_date, date_format)
        
        return start_dt <= end_dt
    
    @staticmethod
    def validate_vndirect_resolution(resolution: str) -> bool:
        valid_resolutions = {'day'}
        return resolution in valid_resolutions if resolution else True
    
    @staticmethod
    def validate_vietstock_resolution(resolution: str) -> bool:
        valid_resolutions = {'1D'}
        return resolution in valid_resolutions if resolution else True
    
    @staticmethod
    def validate_vndirect_stock_index_json_data(data: List[Dict[str, Any]]) -> bool:        
        for entry in data:            
            missing_fields = VistockValidator.VNDIRECT_STOCK_INDEX_REQUIRED_FIELDS - entry.keys()
            if missing_fields:
                return False
            
        return True
    
    @staticmethod
    def validate_dnse_stock_index_json_data(data: List[Dict[str, Any]]) -> bool:
        for entry in data:
            missing_fields = VistockValidator.DNSE_STOCK_INDEX_REQUIRED_FIELDS - entry.keys()
            if missing_fields:
                return False
            
        return True
    
    @staticmethod
    def validate_code(code: str) -> bool:        
        if len(code) == 3:
            return True
        
        return False
    
    @staticmethod
    def validate_enum_value(value: Union[str, Enum], enum_cls: Type[Enum]) -> bool:
        if isinstance(value, enum_cls):
            return True

        if isinstance(value, str):
            if value.upper() in enum_cls.__members__:
                return True
            
            if any(member.value == value for member in enum_cls):
                return True

        return False
    
    @staticmethod
    def validate_year_range(start_year: int, end_year: int) -> bool:
        if start_year > end_year:
            return False
        
        return True
    
    @staticmethod
    def validate_index_code(code: str) -> bool:
        return code in VistockValidator.VNDIRECT_INDEX_CODES if code else True
    
    @staticmethod
    def validate_index_period(period: str) -> bool:
        return period in VistockValidator.VNDIRECT_INDEX_PERIODS if period else True

class VistockNormalizator:
    @staticmethod
    def normalize_enum_value(value: Union[str, Enum], enum_cls: Type[Enum]) -> str:
        if isinstance(value, enum_cls):
            return value.value

        if isinstance(value, str):
            if value.upper() in enum_cls.__members__:
                return enum_cls[value.upper()].value
            for member in enum_cls:
                if member.value == value:
                    return value

        raise ValueError(f"Cannot normalize value '{value}' for enum {enum_cls.__name__}")

class VistockMapper:
    @staticmethod
    def map_english_section(vn_section: str) -> str:
        for section in Vistock24HMoneySectionMapping:
            if section.value == vn_section:
                return section.name.replace('_', ' ').title()
            
        raise ValueError(f'No English mapping found for: {vn_section}')
    
class VistockGenerator:
    QUARTERS = ['03-31', '06-30', '09-30', '12-31']

    @staticmethod
    def generate_annual_dates(start_year: int, end_year: int) -> str:
        dates: List[str] = []

        for year in range(end_year, start_year - 1, -1):
            dates.append(f'{year}-12-31')

        return ','.join(dates)
    
    @staticmethod
    def generate_quarterly_dates(start_year: int, end_year: int) -> str:
        dates: List[str] = []

        for year in range(end_year, start_year - 1, -1):
            for q in reversed(VistockGenerator.QUARTERS):
                dates.append(f'{year}-{q}')

        return ','.join(dates)
    
class VistockConverter:
    @staticmethod
    def convert_utc_to_local(utc_time: str) -> str:
        try:
            utc_time_clean = utc_time.rstrip('Z')

            if '.' in utc_time_clean:
                date_part, ms_part = utc_time_clean.split('.')
                ms_part = (ms_part + '000000')[:6]  # pad to microseconds
                utc_time_clean = f'{date_part}.{ms_part}'
                fmt = '%Y-%m-%dT%H:%M:%S.%f'
            else:
                fmt = '%Y-%m-%dT%H:%M:%S'

            utc_dt = datetime.strptime(utc_time_clean, fmt)
            utc_dt = utc_dt.replace(tzinfo=pytz.utc)

            local_tz = get_localzone()
            local_dt = utc_dt.astimezone(local_tz)

            return local_dt.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
        
        except Exception:
            return utc_time
        
    @staticmethod
    def convert_date_to_timestamp(date: str) -> int:
        dt = datetime.strptime(date, '%Y-%m-%d')
        return int(time.mktime(dt.timetuple()))
    
    @staticmethod
    def convert_timestamp_to_date(timestamp: int) -> str:
        return datetime.fromtimestamp(timestamp).strftime('%Y-%m-%d')

        


from vistock.core.interfaces.ivistockparser import (
    IViStockVnDirectStockIndexParser,
    IVistockVnDirectFundamentalIndexParser,
    IVistockVnDirectFinancialModelsParser,
    IVistockVnDirectFinancialStatementsIndexParser,
    IVistockVnDirectMarketPricesParser,
    IVistockVnDirectChangePricesParser
)
from vistock.core.constants import DEFAULT_VNDIRECT_DOMAIN
from vistock.core.enums import VistockVnDirectFinancialModelsCategory
from vistock.core.utils import VistockValidator, VistockGenerator
from urllib.parse import urlencode
from datetime import datetime
from typing import List, Dict, Any

class VistockVnDirectStockIndexParser(IViStockVnDirectStockIndexParser):
    def __init__(self):
        self._domain = DEFAULT_VNDIRECT_DOMAIN

    def parse_url_path(
        self,
        code: str,
        start_date: str = '2012-01-01',
        end_date: str = datetime.now().strftime('%Y-%m-%d'),
        limit: int = 1
    ) -> str:
        if not VistockValidator.validate_code(code=code):
            raise ValueError(
                'Invalid code: "code" must be a non-empty alphanumeric string with exactly 3 characters representing the stock code. Please ensure that the code is specified correctly.'
            )

        if limit < 0:
            raise ValueError(
                'Invalid limit: "limit" must be a positive integer greater than zero to ensure proper pagination and data retrieval.'
            )
        
        if limit == 0:
            limit += 1

        query_parts = [f'code:{code}']

        if not VistockValidator.validate_date_range(start_date=start_date, end_date=end_date):
            raise ValueError(
                'Invalid date range: "start_date" must be earlier than "end_date". Please ensure that the start date precedes the end date to maintain a valid chronological order.'
            )

        if not VistockValidator.validate_date_format(date_str=start_date):
            raise ValueError(
                f'Invalid start_date format: "{start_date}". Please use "YYYY-MM-DD".'
            )
        query_parts.append(f'date:gte:{start_date}')

        if not VistockValidator.validate_date_format(date_str=end_date):
            raise ValueError(
                f'Invalid end_date format: "{end_date}". Please use "YYYY-MM-DD".'
            )
        query_parts.append(f'date:lte:{end_date}')

        q_param = '~'.join(query_parts)

        query_params : Dict[str, Any] = {
            'sort': 'date',
            'q': q_param,
            'size': limit,
            'page': 1
        }

        return f'?{urlencode(query_params)}'
            
class VistockVnDirectFundamentalIndexParser(IVistockVnDirectFundamentalIndexParser):
    def __init__(self):
        self._domain = DEFAULT_VNDIRECT_DOMAIN
        self._url_templates = [
            '?filter=ratioCode:MARKETCAP,NMVOLUME_AVG_CR_10D,PRICE_HIGHEST_CR_52W,PRICE_LOWEST_CR_52W,OUTSTANDING_SHARES,FREEFLOAT,BETA,PRICE_TO_EARNINGS,PRICE_TO_BOOK,DIVIDEND_YIELD,BVPS_CR,&where=code:{code}&order=reportDate&fields=ratioCode,value',
            '?filter=ratioCode:ROAE_TR_AVG5Q,ROAA_TR_AVG5Q,EPS_TR,&where=code:{code}&order=reportDate&fields=ratioCode,value'
        ]

    def parse_url_path(self, code: str) -> List[str]:
        if not VistockValidator.validate_code(code=code):
            raise ValueError(
                'Invalid code: "code" must be a non-empty alphanumeric string with exactly 3 characters representing the stock code. Please ensure that the code is specified correctly.'
            )
        
        return [template.format(code=code) for template in self._url_templates]
    
class VistockVnDirectFinancialModelsParser(IVistockVnDirectFinancialModelsParser):
    def __init__(self):
        self._domain = DEFAULT_VNDIRECT_DOMAIN
        self._url_template = '?sort=displayOrder:asc&q=codeList:{code}~modelType:{model_type_code}~note:TT199/2014/TT-BTC,TT334/2016/TT-BTC,TT49/2014/TT-NHNN,TT202/2014/TT-BTC~displayLevel:0,1,2,3&size={limit}'

    def parse_url_path(
        self, 
        code: str, 
        model_type_code: str, 
        limit: int = 2000
    ) -> List[str]:
        if not VistockValidator.validate_code(code=code):
            raise ValueError(
                'Invalid code: "code" must be a non-empty alphanumeric string with exactly 3 characters representing the stock code. Please ensure that the code is specified correctly.'
            )
        
        urls: List[str] = []

        if model_type_code == 'all':
            for category in VistockVnDirectFinancialModelsCategory:
                if category.value == 'all':
                    continue

                url = self._url_template.format(
                    code=code,
                    model_type_code=category.value,
                    limit=limit
                )
                urls.append(f'{url}')

            return urls
        else:
            url = self._url_template.format(
                code=code,
                model_type_code=model_type_code,
                limit=limit
            )
            urls.append(f'{url}')

            return urls
    
class VistockVnDirectFinancialStatementsIndexParser(IVistockVnDirectFinancialStatementsIndexParser):
    def __init__(self):
        self._domain = DEFAULT_VNDIRECT_DOMAIN
        self._url_template = '?q=code:{code}~reportType:{report_type_code}~modelType:{model_type_code}~fiscalDate:{fiscal_date}&sort=fiscalDate&size={limit}'

    def parse_url_path(
        self,
        code: str,
        start_year: int = 2000,
        end_year: int = datetime.now().year,
        report_type_code: str = 'ANNUAL',
        model_type_code: str = 'all',
        limit: int = 10000
    ) -> List[str]:
        if not VistockValidator.validate_code(code=code):
            raise ValueError(
                'Invalid code: "code" must be a non-empty alphanumeric string with exactly 3 characters representing the stock code. Please ensure that the code is specified correctly.'
            )
        
        if not VistockValidator.validate_year_range(start_year=start_year, end_year=end_year):
            raise ValueError(
                'Invalid year range: "start_year" must be earlier than or equal to "end_year". Please ensure that the start year precedes or is equal to the end year to maintain a valid chronological order.'
            )
        
        if limit < 0:
            raise ValueError(
                'Invalid limit: "limit" must be a positive integer greater than zero to ensure proper pagination and data retrieval.'
            )
        
        if limit == 0:
            limit += 1
        
        fiscal_date = VistockGenerator.generate_annual_dates(start_year=start_year, end_year=end_year)
        
        if report_type_code == 'QUARTER':
            fiscal_date = VistockGenerator.generate_quarterly_dates(start_year=start_year, end_year=end_year)

        urls: List[str] = []

        if model_type_code == 'all':
            for category in VistockVnDirectFinancialModelsCategory:
                if category.value == 'all':
                    continue

                url = self._url_template.format(
                    code=code,
                    report_type_code=report_type_code,
                    model_type_code=category.value,
                    fiscal_date=fiscal_date,
                    limit=limit
                )
                urls.append(f'{url}')

            return urls
        else:
            url = self._url_template.format(
                code=code,
                report_type_code=report_type_code,
                model_type_code=model_type_code,
                fiscal_date=fiscal_date,
                limit=limit
            )
            urls.append(f'{url}')

            return urls

class VistockVnDirectMarketPricesParser(IVistockVnDirectMarketPricesParser):
    def __init__(self):
        self.domain = DEFAULT_VNDIRECT_DOMAIN
        self._url_template = '?sort=date:{order}&size={limit}&q=code:{code}~date:gte:{start_date}'

    def parse_url_path(
        self,
        code: str,
        start_date: str = '2012-01-01',
        ascending: bool = True,
        limit: int = 1
    ) -> str:
        if not VistockValidator.validate_index_code(code=code):
            raise ValueError(
                f'Invalid index code: "{code}". Must be one of the supported codes: VNINDEX, HNX, UPCOM, VN30, VN30F1M.'
            )
        
        if not VistockValidator.validate_date_format(date_str=start_date):
            raise ValueError(
                f'Invalid start_date format: "{start_date}". Please use "YYYY-MM-DD".'
            )
        
        if limit < 0:
            raise ValueError(
                'Invalid limit: "limit" must be a positive integer greater than zero to ensure proper pagination and data retrieval.'
            )
        
        if limit == 0:
            limit += 1

        if not ascending:
            order = 'desc'

        order = 'asc'

        return self._url_template.format(
            code=code,
            start_date=start_date,
            order=order,
            limit=limit
        )
    
class VistockVnDirectChangePricesParser(IVistockVnDirectChangePricesParser):
    def __init__(self):
        self._domain = DEFAULT_VNDIRECT_DOMAIN
        self._url_template = '?q=code:{code}~period:{period}'

    def parse_url_path(
        self,
        code: str,
        period: str
    ) -> str:
        return self._url_template.format(
            code=code,
            period=period
        )


from vistock.core.enums import (
    Vistock24HMoneyIndustryCategory,
    Vistock24HMoneyFloorCategory,
    Vistock24HMoneyCompanyCategory,
    Vistock24HMoneyLetterCategory
)
from typing import List, Dict, Union, Any, Protocol
from datetime import datetime, timezone

class IViStockVnDirectStockIndexParser(Protocol):
    def parse_url_path(
        self,
        code: str,
        start_date: str = '2012-01-01',
        end_date: str = datetime.now().strftime('%Y-%m-%d'),
        limit: int = 1
    ) -> str:
        ...

class IVistockVnDirectFundamentalIndexParser(Protocol):
    def parse_url_path(
        self,
        code: str
    ) -> List[str]:
        ...

class IVistockVnDirectFinancialModelsParser(Protocol):
    def parse_url_path(
        self,
        code: str,
        model_type_code: str,
        limit: int = 2000
    ) -> List[str]:
        ...

class IVistockVnDirectFinancialStatementsIndexParser(Protocol):
    def parse_url_path(
        self,
        code: str,
        start_year: int = 2000,
        end_year: int = datetime.now().year,
        report_type_code: str = 'ANNUAL',
        model_type_code: str = 'all',
        limit: int = 10000
    ) -> List[str]:
        ...

class IVistockVnDirectMarketPricesParser(Protocol):
    def parse_url_path(
        self,
        code: str,
        start_date: str = '2012-01-01',
        ascending: bool = True,
        limit: int = 1
    ) -> str:
        ...

class IVistockVnDirectChangePricesParser(Protocol):
    def parse_url_path(
        self,
        code: str,
        period: str
    ) -> str:
        ...

class IVistock24HMoneyStockSectionParser(Protocol):
    def parse_url_path(
        self,
        industry_code: Union[Vistock24HMoneyIndustryCategory, str] = 'all',
        floor_code: Union[Vistock24HMoneyFloorCategory, str] = 'all',
        company_code: Union[Vistock24HMoneyCompanyCategory, str] = 'all',
        letter_code: Union[Vistock24HMoneyLetterCategory, str] = 'all',
        limit: int = 2000
    ) -> str:
        ...

class IVistockVietstockStockIndexParser(Protocol):
    def parse_url_path(
        self,
        code: str,
        resolution: str,
        start_date: str = '2000-01-01',
        end_date: str = datetime.now().strftime('%Y-%m-%d')
    ) -> str:
        ...

class IVistockDNSEStockIndexParser(Protocol):
    def parse_payload(
        self,
        code: str,
        current_datetime: str = datetime.now(timezone.utc).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'
    ) -> Dict[str, Any]:
        ...


DEFAULT_VNDIRECT_DOMAIN = 'vndirect.com.vn'
DEFAULT_VNDIRECT_STOCK_INDEX_BASE_URL = 'https://api-finfo.vndirect.com.vn/v4/stock_prices'
DEFAULT_VNDIRECT_FUNDAMENTAL_INDEX_BASE_URL = 'https://api-finfo.vndirect.com.vn/v4/ratios/latest'
DEFAULT_VNDIRECT_FINANCIAL_MODELS_BASE_URL = 'https://api-finfo.vndirect.com.vn/v4/financial_models'
DEFAULT_VNDIRECT_FINANCIAL_STATEMENTS_BASE_URL = 'https://api-finfo.vndirect.com.vn/v4/financial_statements'
DEFAULT_VNDIRECT_MARKET_PRICES_URL = 'https://api-finfo.vndirect.com.vn/v4/vnmarket_prices'
DEFAULT_VNDIRECT_CHANGE_PRICES_URL = 'https://api-finfo.vndirect.com.vn/v4/change_prices'
DEFAULT_VNDIRECT_HEADERS = {
    'Accept': 'application/json',
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/137.0.0.0 Safari/537.36',
    'Origin': 'https://dstock.vndirect.com.vn',
    'Referer': 'https://dstock.vndirect.com.vn/'
}

DEFAULT_24HMONEY_DOMAIN = '24hmoney.vn'
DEFAULT_24HMONEY_BASE_URL = 'https://api-finance-t19.24hmoney.vn/v1/ios/company/az'
DEFAULT_24HMONEY_HEADERS = {
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/127.0.0.0 Safari/537.36',
    'Accept': 'application/json, text/plain, */*',
    'Accept-Encoding': 'gzip, deflate, br, zstd',
    'Accept-Language': 'en-US,en;q=0.9,vi;q=0.8',
    'Origin': 'https://24hmoney.vn',
    'Referer': 'https://24hmoney.vn/',
    'Sec-Ch-Ua': '"Not)A;Brand";v="99", "Google Chrome";v="127", "Chromium";v="127"',
    'Sec-Ch-Ua-Mobile': '?0',
    'Sec-Ch-Ua-Platform': '"Windows"',
    'Sec-Fetch-Dest': 'empty',
    'Sec-Fetch-Mode': 'cors',
    'Sec-Fetch-Site': 'same-site',
}

DEFAULT_VIETSTOCK_DOMAIN = 'api.vietstock.vn'
DEFAULT_VIETSTOCK_STOCK_INDEX_BASE_URL = 'https://api.vietstock.vn/tvnew/history'
DEFAULT_VIETSTOCK_STOCK_INDEX_HEADERS = {
    'Accept': '*/*',
    'Accept-Encoding': 'gzip, deflate, br, zstd',
    'Accept-Language': 'en-US,en;q=0.9,vi;q=0.8',
    'Connection': 'keep-alive',
    'Host': 'api.vietstock.vn',
    'Origin': 'https://stockchart.vietstock.vn',
    'Referer': 'https://stockchart.vietstock.vn/',
    'Sec-Ch-Ua': '"Google Chrome";v="131", "Chromium";v="131", "Not_A Brand";v="24"',
    'Sec-Ch-Ua-Mobile': '?0',
    'Sec-Ch-Ua-Platform': '"Windows"',
    'Sec-Fetch-Dest': 'empty',
    'Sec-Fetch-Mode': 'cors',
    'Sec-Fetch-Site': 'same-site',
    'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/137.0.0.0 Safari/537.36'
}
DEFAULT_DNSE_DOMAIN = 'api.dnse.com.vn'
DEFAULT_DNSE_STOCK_INDEX_BASE_URL = 'https://api.dnse.com.vn/price-api/query'
DEFAULT_DNSE_STOCK_INDEX_HEADERS = {
    "Accept": "*/*",
    "Accept-Encoding": "gzip, deflate, br, zstd",
    "Accept-Language": "vi,en-US;q=0.9,en;q=0.8",
    "Content-Type": "application/json",
    "Origin": "https://banggia.dnse.com.vn",
    "Referer": "https://banggia.dnse.com.vn/",
    "Sec-Ch-Ua": '"Google Chrome";v="137", "Chromium";v="137", "Not/A)Brand";v="24"',
    "Sec-Ch-Ua-Mobile": "?0",
    "Sec-Ch-Ua-Platform": '"Windows"',
    "Sec-Fetch-Dest": "empty",
    "Sec-Fetch-Mode": "cors",
    "Sec-Fetch-Site": "same-site",
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/137.0.0.0 Safari/537.36"
}

DEFAULT_TIMEOUT = 300.0
DEFAULT_TIMEOUT_CONNECT = 150.0


from typing import Dict, Any

class IVistockVnDirectStockIndexScraper:
    def fetch(self, url: str) -> Dict[str, Any]:
        ...

class AsyncIVistockVnDirectStockIndexScraper:
    async def async_fetch(self, url: str) -> Dict[str, Any]:
        ...

class IVistockVnDirectFundamentalIndexScraper:
    def fetch(self, url: str) -> Dict[str, Any]:
        ...

class AsyncIVistockVnDirectFundamentalIndexScraper:
    async def async_fetch(self, url: str) -> Dict[str, Any]:
        ...

class IVistockVnDirectFinancialModelsScraper:
    def fetch(self, url: str) -> Dict[str, Any]:
        ...

class AsyncIVistockVnDirectFinancialModelsScraper:
    async def async_fetch(self, url: str) -> Dict[str, Any]:
        ...

class IVistockVnDirectFinancialStatementsIndexScraper:
    def fetch(self, url: str) -> Dict[str, Any]:
        ...

class AsyncIVistockVnDirectFinancialStatementsIndexScraper:
    async def async_fetch(self, url: str) -> Dict[str, Any]:
        ...

class IVistockVnDirectMarketPricesScraper:
    def fetch(self, url: str) -> Dict[str, Any]:
        ...

class AsyncIVistockVnDirectMarketPricesScraper:
    async def async_fetch(self, url: str) -> Dict[str, Any]:
        ...

class IVistockVnDirectChangePricesScraper:
    def fetch(self, url: str) -> Dict[str, Any]:
        ...

class AsyncIVistockVnDirectChangePricesScraper:
    async def async_fetch(self, url: str) -> Dict[str, Any]:
        ...

class IVistock24HMoneyStockSectionScraper:
    def fetch(self, url: str) -> Dict[str, Any]:
        ...

class AsyncIVistock24HMoneyStockSectionScraper:
    async def async_fetch(self, url: str) -> Dict[str, Any]:
        ...

class IVistockVietstockStockIndexScraper:
    def fetch(self, url: str) -> Dict[str, Any]:
        ...

class AsyncIVistockVietstockStockIndexScraper:
    async def async_fetch(self, url: str) -> Dict[str, Any]:
        ...

class IVistockDNSEStockIndexScraper:
    def fetch(self, url: str, payload: Dict[str, Any]) -> Dict[str, Any]:
        ...

class AsyncIVistockDNSEStockIndexScraper:
    async def async_fetch(self, url: str, payload: Dict[str, Any]) -> Dict[str, Any]:
        ...


from vistock.core.constants import (
    DEFAULT_DNSE_DOMAIN,
    DEFAULT_DNSE_STOCK_INDEX_BASE_URL,
    DEFAULT_DNSE_STOCK_INDEX_HEADERS,
    DEFAULT_TIMEOUT, 
    DEFAULT_TIMEOUT_CONNECT
)
from vistock.core.interfaces.ivistockscraper import (
    IVistockDNSEStockIndexScraper,
    AsyncIVistockDNSEStockIndexScraper
)
from vistock.core.utils import VistockValidator
from typing import Dict, Any
import tenacity
import httpx

class VistockDNSEStockIndexScraper(IVistockDNSEStockIndexScraper, AsyncIVistockDNSEStockIndexScraper):
    def __init__(self, **kwargs: Any) -> None:
        self._base_url = DEFAULT_DNSE_STOCK_INDEX_BASE_URL
        self._domain = DEFAULT_DNSE_DOMAIN

        timeout = kwargs.get('timeout', DEFAULT_TIMEOUT)
        timeout_connect = kwargs.get('timeout_connect', DEFAULT_TIMEOUT_CONNECT)

        if not isinstance(timeout, (int, float)) or not isinstance(timeout_connect, (int, float)):
            raise ValueError(
                'Invalid timeout configuration: both "timeout" and "timeout_connect" must be numeric types (int or float).'
            )
        
        if timeout <= 0 or timeout_connect <= 0:
            raise ValueError(
                'Invalid timeout configuration: both "timeout" and "timeout_connect" must be greater than zero to ensure proper request handling.'
            )
        
        if timeout < timeout_connect:
            raise ValueError(
                'Invalid timeout configuration: the overall "timeout" value must be greater than or equal to "timeout_connect" to avoid premature termination.'
            )
        
        self._timeout = httpx.Timeout(
            timeout=timeout,
            connect=timeout_connect
        )

        headers: Dict[str, Any] = kwargs.get('headers', DEFAULT_DNSE_STOCK_INDEX_HEADERS)

        self._headers = headers

    @tenacity.retry(
        wait=tenacity.wait_exponential(multiplier=1, min=2, max=10),
        stop=tenacity.stop_after_attempt(5),
        reraise=True,
        retry=tenacity.retry_if_exception_type(httpx.HTTPStatusError)
    )
    def fetch(self, url: str, payload: Dict[str, Any]) -> Dict[str, Any]:
        if not VistockValidator.validate_url_with_domain(url, self._domain):
            raise ValueError(
                f'Invalid URL: "{url}". The URL must belong to the expected domain "{self._domain}" to ensure proper validation and access control.'
            ) 
                
        with httpx.Client(timeout=self._timeout) as client:
            response = client.post(url, headers=self._headers, json=payload)
            response.raise_for_status()

        return response.json()
    
    @tenacity.retry(
        wait=tenacity.wait_exponential(multiplier=1, min=2, max=10),
        stop=tenacity.stop_after_attempt(5),
        reraise=True,
        retry=tenacity.retry_if_exception_type(httpx.HTTPStatusError)
    )
    async def async_fetch(self, url: str, payload: Dict[str, Any]) -> Dict[str, Any]:
        if not VistockValidator.validate_url_with_domain(url, self._domain):
            raise ValueError(
                f'Invalid URL: "{url}". The provided URL must belong to the expected domain "{self._domain}" '
                "to ensure source integrity and proper routing within the system."
            )
        
        async with httpx.AsyncClient(timeout=self._timeout) as client:
            response = await client.post(url, headers=self._headers, json=payload)
            response.raise_for_status()

        return response.json()


from vistock.core.constants import (
    DEFAULT_VNDIRECT_STOCK_INDEX_BASE_URL,
    DEFAULT_VNDIRECT_FUNDAMENTAL_INDEX_BASE_URL, 
    DEFAULT_VNDIRECT_FINANCIAL_MODELS_BASE_URL,
    DEFAULT_VNDIRECT_FINANCIAL_STATEMENTS_BASE_URL,
    DEFAULT_VNDIRECT_MARKET_PRICES_URL,
    DEFAULT_VNDIRECT_CHANGE_PRICES_URL,
    DEFAULT_VNDIRECT_DOMAIN, 
    DEFAULT_TIMEOUT
)
from vistock.core.models import (
    StandardVnDirectStockIndexSearch,
    AdvancedVnDirectStockIndexSearch,
    StandardVnDirectFinancialModelSearch,
    StandardVnDirectStockIndexSearchResults,
    AdvancedVnDirectStockIndexSearchResults,
    StandardVnDirectFundamentalIndexSearchResults,
    StandardVnDirectFinancialModelSearchResults,
    StandardVnDirectFinancialStatementsIndex,
    StandardVnDirectFinancialStatementsIndexSearchResults,
    StandardVnDirectMarketPricesSearch,
    StandardVnDirectMarketPricesSearchResults,
    StandardVnDirectChangePricesSearch,
    StandardVnDirectChangePricesSearchResults
)
from vistock.core.enums import (
    VistockVnDirectFinancialModelsCategory,
    VistockVnDirectReportTypeCategory,
    VistockVnDirectIndexCodeMapping,
    VistockVnDirectChangePricePeriodMapping
)
from vistock.core.interfaces.ivistocksearch import (
    IVistockVnDirectStockIndexSearch,
    AsyncIVistockVnDirectStockIndexSearch,
    IVistockVnDirectFundamentalIndexSearch,
    AsyncIVistockVnDirectFundamentalIndexSearch,
    IVistockVnDirectFinancialModelsSearch,
    AsyncIVistockVnDirectFinancialModelsSearch,
    IVistockVnDirectFinancialStatementsIndexSearch,
    AsyncIVistockVnDirectFinancialStatementsIndexSearch,
    IVistockVnDirectMarketPricesSearch,
    AsyncIVistockVnDirectMarketPricesSearch,
    IVistockVnDirectChangePricesSearch,
    AsyncIVistockVnDirectChangePricesSearch
)
from vistock.modules.vndirect.scrapers import (
    VistockVnDirectStockIndexScraper,
    VistockVnDirectFundamentalIndexScraper,
    VistockVnDirectFinancialModelsScraper,
    VistockVnDirectFinancialStatementsIndexScraper,
    VistockVnDirectMarketPricesScraper,
    VistockVnDirectChangePricesScraper
)
from vistock.modules.vndirect.parsers import (
    VistockVnDirectStockIndexParser,
    VistockVnDirectFundamentalIndexParser,
    VistockVnDirectFinancialModelsParser,
    VistockVnDirectFinancialStatementsIndexParser,
    VistockVnDirectMarketPricesParser,
    VistockVnDirectChangePricesParser
)
from vistock.core.utils import VistockValidator, VistockNormalizator
from typing import List, Dict, Tuple, Union, Literal, Any
from datetime import datetime
import asyncio

class VistockVnDirectStockIndexSearch(IVistockVnDirectStockIndexSearch, AsyncIVistockVnDirectStockIndexSearch):
    def __init__(self, timeout: float = DEFAULT_TIMEOUT, **kwargs: Any) -> None:
        if timeout <= 0:
            raise ValueError(
                'Invalid configuration: "timeout" must be a strictly positive integer value representing the maximum allowable wait time for the operation.'
            )
        self._timeout = timeout

        if 'semaphore_limit' in kwargs and (not isinstance(kwargs['semaphore_limit'], int) or kwargs['semaphore_limit'] <= 0):
            raise ValueError(
                'Invalid configuration: "semaphore_limit" must be a positive integer, indicating the maximum number of concurrent asynchronous operations permitted.'
            )

        self._semaphore_limit = kwargs.get('semaphore_limit', 5)
        self._base_url = DEFAULT_VNDIRECT_STOCK_INDEX_BASE_URL
        self._domain = DEFAULT_VNDIRECT_DOMAIN
        self._scraper = VistockVnDirectStockIndexScraper()
        self._parser = VistockVnDirectStockIndexParser()
        self._semaphore = asyncio.Semaphore(self._semaphore_limit)

    @property
    def timeout(self) -> float:
        return self._timeout
    
    @timeout.setter
    def timeout(self, value: int) -> None:
        if value <= 0:
            raise ValueError(
                'Invalid value: "timeout" must be a positive integer greater than zero.'
            )
        self._timeout = value

    def search(
        self,
        code: str,
        start_date: str = '2012-01-01',
        end_date: str = datetime.now().strftime('%Y-%m-%d'),
        resolution: Literal['day'] = 'day',
        advanced: bool = True,
        ascending: bool = False
    ) -> Union[StandardVnDirectStockIndexSearchResults, AdvancedVnDirectStockIndexSearchResults]:
        if not VistockValidator.validate_vndirect_resolution(resolution):
            raise ValueError(
                'Invalid resolution: "resolution" must be one of the following values: "day", "week", "month", or "year". Please ensure that the resolution is specified correctly.'
            )

        initial_url = f'{self._base_url}{self._parser.parse_url_path(code=code, start_date=start_date, end_date=end_date)}'
        total_elements = self._scraper.fetch(url=initial_url).get('totalElements', 0)

        url = f'{self._base_url}{self._parser.parse_url_path(code=code, start_date=start_date, end_date=end_date, limit=total_elements)}'
        data: List[Dict[str, Any]] = self._scraper.fetch(url=url).get('data', [])
        
        if not data:
            raise ValueError(
                'No data found for the given parameters. Please check the code, start date, and end date to ensure they are correct and that data exists for the specified range.'
            )
        
        if not VistockValidator.validate_vndirect_stock_index_json_data(data):
            raise ValueError(
                'Invalid data format: The fetched data does not conform to the expected JSON structure. Please ensure that the API response is valid and contains the necessary fields.'
            )
        
        data.sort(key=lambda x: x.get('date', ''), reverse=not ascending)

        if advanced:
            transformed_data: List[Dict[str, Any]] = [
                {
                    'standard': {
                        'code': item.get('code', ''),
                        'date': item.get('date', ''),
                        'time': item.get('time', ''),
                        'tfloor': item.get('floor', ''),
                        'type': item.get('type', ''),
                        'mopen': item.get('adOpen', 0.0),
                        'mhigh': item.get('adHigh', 0.0),
                        'mlow': item.get('adLow', 0.0),
                        'mclose': item.get('adClose', 0.0),
                        'maverage': item.get('adAverage', 0.0),
                        'nmvolume': int(item.get('nmVolume', 0))
                    },
                    'basic': item.get('basicPrice', 0.0),
                    'ceiling': item.get('ceilingPrice', 0.0),
                    'floor': item.get('floorPrice', 0.0),
                    'open': item.get('open', 0.0),
                    'high': item.get('high', 0.0),
                    'low': item.get('low', 0.0),
                    'close': item.get('close', 0.0),
                    'average': item.get('average', 0.0),
                    'nmvalue': item.get('nmValue', 0.0),
                    'ptvolume': item.get('ptVolume', 0.0),
                    'ptvalue': item.get('ptValue', 0.0),
                    'change': item.get('change', 0.0),
                    'mchange': item.get('adChange', 0.0),
                    'pctchange': item.get('pctChange', 0.0)
                }
                for item in data
            ]

            return AdvancedVnDirectStockIndexSearchResults(
                results=[AdvancedVnDirectStockIndexSearch(**item) for item in transformed_data],
                total_results=len(transformed_data)
            )

        return StandardVnDirectStockIndexSearchResults(
            results=[
                StandardVnDirectStockIndexSearch(
                    code=item.get('code', ''),
                    date=item.get('date', ''),
                    time=item.get('time', ''),
                    tfloor=item.get('floor', ''),
                    type=item.get('type', ''),
                    mopen=item.get('adOpen', 0.0),
                    mhigh=item.get('adHigh', 0.0),
                    mlow=item.get('adLow', 0.0),
                    mclose=item.get('adClose', 0.0),
                    maverage=item.get('adAverage', 0.0),
                    nmvolume=int(item.get('nmVolume', 0))
                ) for item in data
            ],
            total_results=len(data)
        ) 

    async def async_search(
        self,
        code: str,
        start_date: str = '2012-01-01',
        end_date: str = datetime.now().strftime('%Y-%m-%d'),
        resolution: Literal['day'] = 'day',
        advanced: bool = True,
        ascending: bool = False
    ) -> Union[StandardVnDirectStockIndexSearchResults, AdvancedVnDirectStockIndexSearchResults]:
        if not VistockValidator.validate_vndirect_resolution(resolution):
            raise ValueError(
                'Invalid resolution: "resolution" must be one of the following values: "day", "week", "month", or "year". Please ensure that the resolution is specified correctly.'
            )

        initial_url = f'{self._base_url}{self._parser.parse_url_path(code=code, start_date=start_date, end_date=end_date)}'
        initial_response = await self._scraper.async_fetch(url=initial_url)
        total_elements = initial_response.get('totalElements', 0)

        url = f'{self._base_url}{self._parser.parse_url_path(code=code, start_date=start_date, end_date=end_date, limit=total_elements)}'
        response = await self._scraper.async_fetch(url=url) 
        data: List[Dict[str, Any]] = response.get('data', [])

        if not data:
            raise ValueError(
                'No data found for the given parameters. Please check the code, start date, and end date to ensure they are correct and that data exists for the specified range.'
            )

        if not VistockValidator.validate_vndirect_stock_index_json_data(data):
            raise ValueError(
                'Invalid data format: The fetched data does not conform to the expected JSON structure. Please ensure that the API response is valid and contains the necessary fields.'
            )

        data.sort(key=lambda x: x.get('date', ''), reverse=not ascending)

        if advanced:
            transformed_data: List[Dict[str, Any]] = [
                {
                    'standard': {
                        'code': item.get('code', ''),
                        'date': item.get('date', ''),
                        'time': item.get('time', ''),
                        'tfloor': item.get('floor', ''),
                        'type': item.get('type', ''),
                        'mopen': item.get('adOpen', 0.0),
                        'mhigh': item.get('adHigh', 0.0),
                        'mlow': item.get('adLow', 0.0),
                        'mclose': item.get('adClose', 0.0),
                        'maverage': item.get('adAverage', 0.0),
                        'nmvolume': int(item.get('nmVolume', 0))
                    },
                    'basic': item.get('basicPrice', 0.0),
                    'ceiling': item.get('ceilingPrice', 0.0),
                    'floor': item.get('floorPrice', 0.0),
                    'open': item.get('open', 0.0),
                    'high': item.get('high', 0.0),
                    'low': item.get('low', 0.0),
                    'close': item.get('close', 0.0),
                    'average': item.get('average', 0.0),
                    'nmvalue': item.get('nmValue', 0.0),
                    'ptvolume': item.get('ptVolume', 0.0),
                    'ptvalue': item.get('ptValue', 0.0),
                    'change': item.get('change', 0.0),
                    'mchange': item.get('adChange', 0.0),
                    'pctchange': item.get('pctChange', 0.0)
                }
                for item in data
            ]

            return AdvancedVnDirectStockIndexSearchResults(
                results=[AdvancedVnDirectStockIndexSearch(**item) for item in transformed_data],
                total_results=len(transformed_data)
            )

        return StandardVnDirectStockIndexSearchResults(
            results=[
                StandardVnDirectStockIndexSearch(
                    code=item.get('code', ''),
                    date=item.get('date', ''),
                    time=item.get('time', ''),
                    tfloor=item.get('floor', ''),
                    type=item.get('type', ''),
                    mopen=item.get('adOpen', 0.0),
                    mhigh=item.get('adHigh', 0.0),
                    mlow=item.get('adLow', 0.0),
                    mclose=item.get('adClose', 0.0),
                    maverage=item.get('adAverage', 0.0),
                    nmvolume=int(item.get('nmVolume', 0))
                ) for item in data
            ],
            total_results=len(data)
        )

class VistockVnDirectFundamentalIndexSearch(IVistockVnDirectFundamentalIndexSearch, AsyncIVistockVnDirectFundamentalIndexSearch):
    def __init__(self, timeout: float = DEFAULT_TIMEOUT, **kwargs: Any) -> None:
        if timeout <= 0:
            raise ValueError(
                'Invalid configuration: "timeout" must be a strictly positive integer value representing the maximum allowable wait time for the operation.'
            )
        self._timeout = timeout

        if 'semaphore_limit' in kwargs and (not isinstance(kwargs['semaphore_limit'], int) or kwargs['semaphore_limit'] <= 0):
            raise ValueError(
                'Invalid configuration: "semaphore_limit" must be a positive integer, indicating the maximum number of concurrent asynchronous operations permitted.'
            )

        self._semaphore_limit = kwargs.get('semaphore_limit', 5)
        self._base_url = DEFAULT_VNDIRECT_FUNDAMENTAL_INDEX_BASE_URL
        self._domain = DEFAULT_VNDIRECT_DOMAIN
        self._scraper = VistockVnDirectFundamentalIndexScraper()
        self._parser = VistockVnDirectFundamentalIndexParser()
        self._semaphore = asyncio.Semaphore(self._semaphore_limit)

    @property
    def timeout(self) -> float:
        return self._timeout
    
    @timeout.setter
    def timeout(self, value: int) -> None:
        if value <= 0:
            raise ValueError(
                'Invalid value: "timeout" must be a positive integer greater than zero.'
            )
        self._timeout = value

    def search(self, code: str) -> StandardVnDirectFundamentalIndexSearchResults:
        results: List[List[Dict[str, Any]]] = []
        
        urls = self._parser.parse_url_path(code=code)
        for url in urls:
            url = f'{self._base_url}{url}'
            data: List[Dict[str, Any]] = self._scraper.fetch(url).get('data', [])

            if not data:
                raise ValueError(
                    'No data found for the given parameters. Please check the code to ensure they are correct.'
                )
            
            results.append(data)

        merged_results = [item for result in results for item in result]

        field_map = {
            'MARKETCAP': 'marketcap',
            'NMVOLUME_AVG_CR_10D': 'nm_volume_avg_cr_10d',
            'PRICE_HIGHEST_CR_52W': 'price_highest_cr_52w',
            'PRICE_LOWEST_CR_52W': 'price_lowest_cr_52w',
            'OUTSTANDING_SHARES': 'outstanding_shares',
            'FREEFLOAT': 'freefloat',
            'BETA': 'beta',
            'PRICE_TO_EARNINGS': 'price_to_earnings',
            'PRICE_TO_BOOK': 'price_to_book',
            'ROAE_TR_AVG5Q': 'roae_tr_avg_5q',
            'ROAA_TR_AVG5Q': 'roaa_tr_avg_5q',
            'DIVIDEND_YIELD': 'dividend_yield',
            'EPS_TR': 'eps_tr',
            'BVPS_CR': 'bvps_cr'
        }

        model_data: Dict[str, float] = {}
        for item in merged_results:
            key = field_map.get(item.get('ratioCode', ''))
            if key:
                model_data[key] = item.get('value', 0.0)

        return StandardVnDirectFundamentalIndexSearchResults(**model_data)

    async def async_search(self, code: str) -> StandardVnDirectFundamentalIndexSearchResults:           
        results: List[List[Dict[str, Any]]] = []
        
        urls = self._parser.parse_url_path(code=code)
        for url in urls:
            url = f'{self._base_url}{url}'
            response = await self._scraper.async_fetch(url)
            data: List[Dict[str, Any]] = response.get('data', [])

            if not data:
                raise ValueError(
                    'No data found for the given parameters. Please check the code to ensure they are correct.'
                )
            
            results.append(data)

        merged_results = [item for result in results for item in result]

        field_map = {
            'MARKETCAP': 'marketcap',
            'NMVOLUME_AVG_CR_10D': 'nm_volume_avg_cr_10d',
            'PRICE_HIGHEST_CR_52W': 'price_highest_cr_52w',
            'PRICE_LOWEST_CR_52W': 'price_lowest_cr_52w',
            'OUTSTANDING_SHARES': 'outstanding_shares',
            'FREEFLOAT': 'freefloat',
            'BETA': 'beta',
            'PRICE_TO_EARNINGS': 'price_to_earnings',
            'PRICE_TO_BOOK': 'price_to_book',
            'ROAE_TR_AVG5Q': 'roae_tr_avg_5q',
            'ROAA_TR_AVG5Q': 'roaa_tr_avg_5q',
            'DIVIDEND_YIELD': 'dividend_yield',
            'EPS_TR': 'eps_tr',
            'BVPS_CR': 'bvps_cr'
        }

        model_data: Dict[str, float] = {}
        for item in merged_results:
            key = field_map.get(item.get('ratioCode', ''))
            if key:
                model_data[key] = item.get('value', 0.0)

        return StandardVnDirectFundamentalIndexSearchResults(**model_data)

class VistockVnDirectFinancialModelsSearch(IVistockVnDirectFinancialModelsSearch, AsyncIVistockVnDirectFinancialModelsSearch):
    def __init__(self, timeout: float = DEFAULT_TIMEOUT, **kwargs: Any) -> None:
        if timeout <= 0:
            raise ValueError(
                'Invalid configuration: "timeout" must be a strictly positive integer value representing the maximum allowable wait time for the operation.'
            )
        self._timeout = timeout

        if 'semaphore_limit' in kwargs and (not isinstance(kwargs['semaphore_limit'], int) or kwargs['semaphore_limit'] <= 0):
            raise ValueError(
                'Invalid configuration: "semaphore_limit" must be a positive integer, indicating the maximum number of concurrent asynchronous operations permitted.'
            )

        self._semaphore_limit = kwargs.get('semaphore_limit', 5)
        self._base_url = DEFAULT_VNDIRECT_FINANCIAL_MODELS_BASE_URL
        self._domain = DEFAULT_VNDIRECT_DOMAIN
        self._scraper = VistockVnDirectFinancialModelsScraper()
        self._parser = VistockVnDirectFinancialModelsParser()
        self._semaphore = asyncio.Semaphore(self._semaphore_limit)

    def search(
        self,
        code: str,
        model_type: Union[VistockVnDirectFinancialModelsCategory, str] = 'all'
    ) -> StandardVnDirectFinancialModelSearchResults:
        if model_type != 'all':
            if not VistockValidator.validate_enum_value(model_type, VistockVnDirectFinancialModelsCategory):
                raise ValueError(f'"{model_type}" is not a recognized model type. Use a valid enum name or code.')
            model_type_code = VistockNormalizator.normalize_enum_value(model_type, VistockVnDirectFinancialModelsCategory)
        else:
            model_type_code = 'all'

        results: List[List[Dict[str, Any]]] = []

        urls = self._parser.parse_url_path(code=code, model_type_code=model_type_code)
        for url in urls:
            url = f'{self._base_url}{url}'
            data: List[Dict[str, Any]] = self._scraper.fetch(url=url).get('data', [])

            if not data:
                raise ValueError(
                    'No data found for the given parameters. Please check the code, and model type to ensure they are correct.'
                )
            
            results.append(data)

        merged_results = [item for result in results for item in result]

        return StandardVnDirectFinancialModelSearchResults(
            results=[
                StandardVnDirectFinancialModelSearch(
                    model_type=item.get('modelType', 0),
                    model_type_name=item.get('modelTypeName', ''),
                    model_vn_desc=item.get('modelVnDesc', ''),
                    model_en_desc=item.get('modelEnDesc', ''),
                    company_form=item.get('companyForm', ''),
                    note=item.get('note', ''),
                    item_code=item.get('itemCode', 0),
                    item_vn_name=item.get('itemVnName', ''),
                    item_en_name=item.get('itemEnName', ''),
                    display_order=item.get('displayOrder', 0),
                    display_level=item.get('displayLevel', 0),
                    form_type=item.get('formType', '')
                ) for item in merged_results
            ],
            total_results=len(merged_results)
        )

    async def async_search(
        self,
        code: str,
        model_type: Union[VistockVnDirectFinancialModelsCategory, str] = 'all'
    ) -> StandardVnDirectFinancialModelSearchResults:
        if model_type != 'all':
            if not VistockValidator.validate_enum_value(model_type, VistockVnDirectFinancialModelsCategory):
                raise ValueError(f'"{model_type}" is not a recognized model type. Use a valid enum name or code.')
            model_type_code = VistockNormalizator.normalize_enum_value(model_type, VistockVnDirectFinancialModelsCategory)
        else:
            model_type_code = 'all'

        results: List[List[Dict[str, Any]]] = []

        urls = self._parser.parse_url_path(code=code, model_type_code=model_type_code)
        for url in urls:
            url = f'{self._base_url}{url}'
            response = await self._scraper.async_fetch(url=url)
            data: List[Dict[str, Any]] = response.get('data', [])

            if not data:
                raise ValueError(
                    'No data found for the given parameters. Please check the code, and model type to ensure they are correct.'
                )
            
            results.append(data)

        merged_results = [item for result in results for item in result]

        return StandardVnDirectFinancialModelSearchResults(
            results=[
                StandardVnDirectFinancialModelSearch(
                    model_type=item.get('modelType', 0),
                    model_type_name=item.get('modelTypeName', ''),
                    model_vn_desc=item.get('modelVnDesc', ''),
                    model_en_desc=item.get('modelEnDesc', ''),
                    company_form=item.get('companyForm', ''),
                    note=item.get('note', ''),
                    item_code=item.get('itemCode', 0),
                    item_vn_name=item.get('itemVnName', ''),
                    item_en_name=item.get('itemEnName', ''),
                    display_order=item.get('displayOrder', 0),
                    display_level=item.get('displayLevel', 0),
                    form_type=item.get('formType', '')
                ) for item in merged_results
            ],
            total_results=len(merged_results)
        )

class VistockVnDirectFinancialStatementsIndexSearch(IVistockVnDirectFinancialStatementsIndexSearch, AsyncIVistockVnDirectFinancialStatementsIndexSearch):
    def __init__(self, timeout: float = DEFAULT_TIMEOUT, **kwargs: Any) -> None:
        if timeout <= 0:
            raise ValueError(
                'Invalid configuration: "timeout" must be a strictly positive integer value representing the maximum allowable wait time for the operation.'
            )
        self._timeout = timeout

        if 'semaphore_limit' in kwargs and (not isinstance(kwargs['semaphore_limit'], int) or kwargs['semaphore_limit'] <= 0):
            raise ValueError(
                'Invalid configuration: "semaphore_limit" must be a positive integer, indicating the maximum number of concurrent asynchronous operations permitted.'
            )

        self._semaphore_limit = kwargs.get('semaphore_limit', 5)
        self._base_url = DEFAULT_VNDIRECT_FINANCIAL_STATEMENTS_BASE_URL
        self._domain = DEFAULT_VNDIRECT_DOMAIN
        self._scraper = VistockVnDirectFinancialStatementsIndexScraper()
        self._parser = VistockVnDirectFinancialStatementsIndexParser()
        self._finanical_models_search = VistockVnDirectFinancialModelsSearch()
        self._semaphore = asyncio.Semaphore(self._semaphore_limit)

    def search(
        self,
        code: str,
        start_year: int = 2000,
        end_year: int = datetime.now().year,
        report_type: Union[VistockVnDirectReportTypeCategory, str] = 'ANNUAL',
        model_type: Union[VistockVnDirectFinancialModelsCategory, str] = 'all'
    ) -> StandardVnDirectFinancialStatementsIndexSearchResults:
        financial_models = self._finanical_models_search.search(
            code=code,
            model_type=model_type
        )

        if report_type != 'ANNUAL':
            if not VistockValidator.validate_enum_value(report_type, VistockVnDirectReportTypeCategory):
                raise ValueError(f'"{report_type}" is not a recognized report type. Use a valid enum name or code.')
            report_type_code = VistockNormalizator.normalize_enum_value(report_type, VistockVnDirectReportTypeCategory)
        else:
            report_type_code = 'all'

        if model_type != 'all':
            if not VistockValidator.validate_enum_value(model_type, VistockVnDirectFinancialModelsCategory):
                raise ValueError(f'"{model_type}" is not a recognized model type. Use a valid enum name or code.')
            model_type_code = VistockNormalizator.normalize_enum_value(model_type, VistockVnDirectFinancialModelsCategory)
        else:
            model_type_code = 'all'     

        results: List[StandardVnDirectFinancialStatementsIndex] = []

        urls = self._parser.parse_url_path(
            code=code,
            start_year=start_year,
            end_year=end_year,
            report_type_code=report_type_code,
            model_type_code=model_type_code
        )

        for url in urls:
            url = f'{self._base_url}{url}'
            data: List[Dict[str, Any]] = self._scraper.fetch(url).get('data', [])

            if not data:
                raise ValueError(
                    'No data found for the given parameters. Please check the code, start year, end year, report type, and model type to ensure they are correct and that data exists for the specified range.'
                )

            for model in financial_models.results:
                for item in data:
                    if (
                        item.get('modelType') == model.model_type and
                        item.get('itemCode') == model.item_code
                    ):
                        index = StandardVnDirectFinancialStatementsIndex(
                            code=item.get('code', ''),
                            model=model,
                            report_type=item.get('reportType', ''),
                            numeric_value=item.get('numericValue', 0.0),
                            fiscal_date=item.get('fiscalDate', ''),
                            created_date=item.get('createdDate', ''),
                            modified_date=item.get('modifiedDate', '')
                        )
                        results.append(index)

        return StandardVnDirectFinancialStatementsIndexSearchResults(
            results=results,
            total_results=len(results)
        )

    async def async_search(
        self,
        code: str,
        start_year: int = 2000,
        end_year: int = datetime.now().year,
        report_type: Union[VistockVnDirectReportTypeCategory, str] = 'ANNUAL',
        model_type: Union[VistockVnDirectFinancialModelsCategory, str] = 'all'
    ) -> StandardVnDirectFinancialStatementsIndexSearchResults:
        financial_models = await self._finanical_models_search.async_search(
            code=code,
            model_type=model_type
        )

        if report_type != 'ANNUAL':
            if not VistockValidator.validate_enum_value(report_type, VistockVnDirectReportTypeCategory):
                raise ValueError(f'"{report_type}" is not a recognized report type. Use a valid enum name or code.')
            report_type_code = VistockNormalizator.normalize_enum_value(report_type, VistockVnDirectReportTypeCategory)
        else:
            report_type_code = 'all'

        if model_type != 'all':
            if not VistockValidator.validate_enum_value(model_type, VistockVnDirectFinancialModelsCategory):
                raise ValueError(f'"{model_type}" is not a recognized model type. Use a valid enum name or code.')
            model_type_code = VistockNormalizator.normalize_enum_value(model_type, VistockVnDirectFinancialModelsCategory)
        else:
            model_type_code = 'all'     

        results: List[StandardVnDirectFinancialStatementsIndex] = []

        urls = self._parser.parse_url_path(
            code=code,
            start_year=start_year,
            end_year=end_year,
            report_type_code=report_type_code,
            model_type_code=model_type_code
        )

        model_lookup: Dict[Tuple[int, int], StandardVnDirectFinancialModelSearch] = {
            (model.model_type, model.item_code): model
            for model in financial_models.results
        }

        for url in urls:
            url = f'{self._base_url}{url}'
            response = await self._scraper.async_fetch(url)
            data: List[Dict[str, Any]] = response.get('data', [])

            if not data:
                raise ValueError(
                'No data found for the given parameters. Please check the code, start year, end year, report type, and model type to ensure they are correct and that data exists for the specified range.'
            )

            for item in data:
                model_key = (item.get('modelType', ''), item.get('itemCode', ''))
                model = model_lookup.get(model_key)
                if model is None:
                    continue

                index = StandardVnDirectFinancialStatementsIndex(
                    code=item.get('code', ''),
                    model=model,
                    report_type=item.get('reportType', ''),
                    numeric_value=item.get('numericValue', 0.0),
                    fiscal_date=item.get('fiscalDate', ''),
                    created_date=item.get('createdDate', ''),
                    modified_date=item.get('modifiedDate', '')
                )
                results.append(index)

        return StandardVnDirectFinancialStatementsIndexSearchResults(
            results=results,
            total_results=len(results)
        )
    
class VistockVnDirectMarketPricesSearch(IVistockVnDirectMarketPricesSearch, AsyncIVistockVnDirectMarketPricesSearch):
    def __init__(self, timeout: float = DEFAULT_TIMEOUT, **kwargs: Any) -> None:
        if timeout <= 0:
            raise ValueError(
                'Invalid configuration: "timeout" must be a strictly positive integer value representing the maximum allowable wait time for the operation.'
            )
        self._timeout = timeout

        if 'semaphore_limit' in kwargs and (not isinstance(kwargs['semaphore_limit'], int) or kwargs['semaphore_limit'] <= 0):
            raise ValueError(
                'Invalid configuration: "semaphore_limit" must be a positive integer, indicating the maximum number of concurrent asynchronous operations permitted.'
            )

        self._semaphore_limit = kwargs.get('semaphore_limit', 5)
        self._base_url = DEFAULT_VNDIRECT_MARKET_PRICES_URL
        self._domain = DEFAULT_VNDIRECT_DOMAIN
        self._scraper = VistockVnDirectMarketPricesScraper()
        self._parser = VistockVnDirectMarketPricesParser()
        self._semaphore = asyncio.Semaphore(self._semaphore_limit)

    @property
    def timeout(self) -> float:
        return self._timeout
    
    @timeout.setter
    def timeout(self, value: int) -> None:
        if value <= 0:
            raise ValueError(
                'Invalid value: "timeout" must be a positive integer greater than zero.'
            )
        self._timeout = value

    def search(
        self,
        code: Union[VistockVnDirectIndexCodeMapping, str],
        start_date: str = '2012-01-01',
        ascending: bool = True
    ) -> StandardVnDirectMarketPricesSearchResults:
        if not VistockValidator.validate_enum_value(code, VistockVnDirectIndexCodeMapping):
            raise ValueError(
                f'Invalid index code: "{code}". Must be one of the supported codes: VNINDEX, HNX, UPCOM, VN30, VN30F1M.'
            )
        
        code = VistockNormalizator.normalize_enum_value(code, VistockVnDirectIndexCodeMapping)

        if code == 'VN30F1M':
            raise ValueError(
                f'The index code "{code}" is not supported for this search operation.'
            )
        
        initial_url = f'{self._base_url}{self._parser.parse_url_path(code=code, start_date=start_date, ascending=ascending)}'
        total_elements = self._scraper.fetch(url=initial_url).get('totalElements', 0)

        url = f'{self._base_url}{self._parser.parse_url_path(code=code, start_date=start_date, ascending=ascending, limit=total_elements)}'
        data: List[Dict[str, Any]] = self._scraper.fetch(url=url).get('data', [])

        if not data:
            raise ValueError(
                'No data found for the given parameters. Please check the code and start date to ensure they are correct and that data exists for the specified range.'
            )
        
        return StandardVnDirectMarketPricesSearchResults(
            results=[
                StandardVnDirectMarketPricesSearch(
                    code=item.get('code', ''),
                    date=item.get('date', ''),
                    time=item.get('time', ''),
                    tfloor=item.get('floor', ''),
                    type=item.get('type', ''),
                    open=item.get('open', 0.0),
                    high=item.get('high', 0.0),
                    low=item.get('low', 0.0),
                    close=item.get('close', 0.0),
                    change=item.get('change', 0.0),
                    pct_change=item.get('pctChange', 0.0),
                    accumulated_volume=item.get('accumulatedVol', 0.0),
                    accumulated_value=item.get('accumulatedVal', 0.0),
                    nmvolume=item.get('nmVolume', 0.0),
                    nmvalue=item.get('nmValue', 0.0),
                    ptvolume=item.get('ptVolume', 0.0),
                    ptvalue=item.get('ptValue', 0.0),
                    advances=item.get('advances', 0.0),
                    declines=item.get('declines', 0.0),
                    no_change=item.get('noChange', 0.0),
                    no_trade=item.get('noTrade', 0.0),
                    ceiling_stocks=item.get('ceilingStocks', 0.0),
                    floor_stocks=item.get('floorStocks', 0.0),
                    val_chg_pct_cr1d=item.get('valChgPctCr1D', 0.0)
                ) for item in data
            ],
            total_results=len(data)
        )
    
    async def async_search(
        self,
        code: Union[VistockVnDirectIndexCodeMapping, str],
        start_date: str = '2012-01-01',
        ascending: bool = True
    ) -> StandardVnDirectMarketPricesSearchResults:
        if not VistockValidator.validate_enum_value(code, VistockVnDirectIndexCodeMapping):
            raise ValueError(
                f'Invalid index code: "{code}". Must be one of the supported codes: VNINDEX, HNX, UPCOM, VN30, VN30F1M.'
            )
        
        code = VistockNormalizator.normalize_enum_value(code, VistockVnDirectIndexCodeMapping)

        if code == 'VN30F1M':
            raise ValueError(
                f'The index code "{code}" is not supported for this search operation.'
            )

        initial_url = f'{self._base_url}{self._parser.parse_url_path(code=code, start_date=start_date, ascending=ascending)}'
        initial_response = await self._scraper.async_fetch(url=initial_url)
        total_elements = initial_response.get('totalElements', 0)

        url = f'{self._base_url}{self._parser.parse_url_path(code=code, start_date=start_date, ascending=ascending, limit=total_elements)}'
        response = await self._scraper.async_fetch(url=url)
        data: List[Dict[str, Any]] = response.get('data', [])

        if not data:
            raise ValueError(
                'No data found for the given parameters. Please check the code and start date to ensure they are correct and that data exists for the specified range.'
            )
        
        return StandardVnDirectMarketPricesSearchResults(
            results=[
                StandardVnDirectMarketPricesSearch(
                    code=item.get('code', ''),
                    date=item.get('date', ''),
                    time=item.get('time', ''),
                    tfloor=item.get('floor', ''),
                    type=item.get('type', ''),
                    open=item.get('open', 0.0),
                    high=item.get('high', 0.0),
                    low=item.get('low', 0.0),
                    close=item.get('close', 0.0),
                    change=item.get('change', 0.0),
                    pct_change=item.get('pctChange', 0.0),
                    accumulated_volume=item.get('accumulatedVol', 0.0),
                    accumulated_value=item.get('accumulatedVal', 0.0),
                    nmvolume=item.get('nmVolume', 0.0),
                    nmvalue=item.get('nmValue', 0.0),
                    ptvolume=item.get('ptVolume', 0.0),
                    ptvalue=item.get('ptValue', 0.0),
                    advances=item.get('advances', 0.0),
                    declines=item.get('declines', 0.0),
                    no_change=item.get('noChange', 0.0),
                    no_trade=item.get('noTrade', 0.0),
                    ceiling_stocks=item.get('ceilingStocks', 0.0),
                    floor_stocks=item.get('floorStocks', 0.0),
                    val_chg_pct_cr1d=item.get('valChgPctCr1D', 0.0)
                ) for item in data
            ],
            total_results=len(data)
        )
    
class VistockVnDirectChangePricesSearch(IVistockVnDirectChangePricesSearch, AsyncIVistockVnDirectChangePricesSearch):
    def __init__(self, timeout: float = DEFAULT_TIMEOUT, **kwargs: Any) -> None:
        if timeout <= 0:
            raise ValueError(
                'Invalid configuration: "timeout" must be a strictly positive integer value representing the maximum allowable wait time for the operation.'
            )
        self._timeout = timeout

        if 'semaphore_limit' in kwargs and (not isinstance(kwargs['semaphore_limit'], int) or kwargs['semaphore_limit'] <= 0):
            raise ValueError(
                'Invalid configuration: "semaphore_limit" must be a positive integer, indicating the maximum number of concurrent asynchronous operations permitted.'
            )

        self._semaphore_limit = kwargs.get('semaphore_limit', 5)
        self._base_url = DEFAULT_VNDIRECT_CHANGE_PRICES_URL
        self._domain = DEFAULT_VNDIRECT_DOMAIN
        self._scraper = VistockVnDirectChangePricesScraper()
        self._parser = VistockVnDirectChangePricesParser()
        self._semaphore = asyncio.Semaphore(self._semaphore_limit)

    @property
    def timeout(self) -> float:
        return self._timeout
    
    @timeout.setter
    def timeout(self, value: int) -> None:
        if value <= 0:
            raise ValueError(
                'Invalid value: "timeout" must be a positive integer greater than zero.'
            )
        self._timeout = value

    def search(
        self,
        code: Union[VistockVnDirectIndexCodeMapping, str] = 'VNINDEX,HNX,UPCOM,VN30,VN30F1M',
        period: Union[VistockVnDirectChangePricePeriodMapping, str] = '1D'
    ) -> StandardVnDirectChangePricesSearchResults:
        if isinstance(code, str):
            index_codes = code.split(',')
        else:
            index_codes = [code]

        normalized_index_codes: List[str] = []
        for index_code in index_codes:
            if not VistockValidator.validate_enum_value(index_code, VistockVnDirectIndexCodeMapping):
                raise ValueError(
                f'Invalid index code: "{code}". Must be one of the supported codes: VNINDEX, HNX, UPCOM, VN30, VN30F1M.'
            )
            normalized_index_codes.append(
                VistockNormalizator.normalize_enum_value(index_code, VistockVnDirectIndexCodeMapping)
            )

        code = ','.join(normalized_index_codes)

        if not VistockValidator.validate_enum_value(period, VistockVnDirectChangePricePeriodMapping):
            raise ValueError(
                f'Invalid index period: "{period}". Must be one of the supported periods: 1D, MTD, QTD, YTD, 5D, 1M, 3M, 6M, 1Y.'
            )
        
        period = VistockNormalizator.normalize_enum_value(period, VistockVnDirectChangePricePeriodMapping)

        url = f'{self._base_url}{self._parser.parse_url_path(code=code, period=period)}'
        data: List[Dict[str, Any]] = self._scraper.fetch(url=url).get('data', [])

        if not data:
            raise ValueError(
                'No data found for the given parameters. Please check the code and period to ensure they are correct and that data exists for the specified range.'
            )
        
        return StandardVnDirectChangePricesSearchResults(
            results=[
                StandardVnDirectChangePricesSearch(
                    code=item.get('code', ''),
                    name=item.get('name', ''),
                    type=item.get('type', ''),
                    period=item.get('period', ''),
                    price=item.get('price', 0.0),
                    bop_price=item.get('bopPrice', 0.0),
                    change=item.get('change', 0.0),
                    pct_change=item.get('changePct', 0.0),
                    last_updated=item.get('lastUpdated', '')
                ) for item in data
            ],
            total_results=len(data)
        )
        
    async def async_search(
        self,
        code: Union[VistockVnDirectIndexCodeMapping, str] = 'VNINDEX,HNX,UPCOM,VN30,VN30F1M',
        period: Union[VistockVnDirectChangePricePeriodMapping, str] = '1D'
    ) -> StandardVnDirectChangePricesSearchResults:
        if isinstance(code, str):
            index_codes = code.split(',')
        else:
            index_codes = [code]

        normalized_index_codes: List[str] = []
        for index_code in index_codes:
            if not VistockValidator.validate_enum_value(index_code, VistockVnDirectIndexCodeMapping):
                raise ValueError(
                f'Invalid index code: "{code}". Must be one of the supported codes: VNINDEX, HNX, UPCOM, VN30, VN30F1M.'
            )
            normalized_index_codes.append(
                VistockNormalizator.normalize_enum_value(index_code, VistockVnDirectIndexCodeMapping)
            )

        code = ','.join(normalized_index_codes)

        if not VistockValidator.validate_enum_value(period, VistockVnDirectChangePricePeriodMapping):
            raise ValueError(
                f'Invalid index period: "{period}". Must be one of the supported periods: 1D, MTD, QTD, YTD, 5D, 1M, 3M, 6M, 1Y.'
            )
        
        period = VistockNormalizator.normalize_enum_value(period, VistockVnDirectChangePricePeriodMapping)

        url = f'{self._base_url}{self._parser.parse_url_path(code=code, period=period)}'
        response = await self._scraper.async_fetch(url=url)
        data: List[Dict[str, Any]] = response.get('data', [])

        if not data:
            raise ValueError(
                'No data found for the given parameters. Please check the code and period to ensure they are correct and that data exists for the specified range.'
            )
        
        return StandardVnDirectChangePricesSearchResults(
            results=[
                StandardVnDirectChangePricesSearch(
                    code=item.get('code', ''),
                    name=item.get('name', ''),
                    type=item.get('type', ''),
                    period=item.get('period', ''),
                    price=item.get('price', 0.0),
                    bop_price=item.get('bopPrice', 0.0),
                    change=item.get('change', 0.0),
                    pct_change=item.get('changePct', 0.0),
                    last_updated=item.get('lastUpdated', '')
                ) for item in data
            ],
            total_results=len(data)
        )


from vistock.core.constants import (
    DEFAULT_DNSE_DOMAIN,
    DEFAULT_DNSE_STOCK_INDEX_BASE_URL,
    DEFAULT_TIMEOUT
)
from vistock.core.models import (
    StandardDNSEStockIndexSearch,
    StandardDNSEStockIndexSearchResults
)
from vistock.modules.dnse.scrapers import VistockDNSEStockIndexScraper
from vistock.modules.dnse.parsers import VistockDNSEStockIndexParser
from vistock.core.interfaces.ivistocksearch import (
    IVistockDNSEStockIndexSearch,
    AsyncIVistockDNSEStockIndexSearch
)
from vistock.core.utils import VistockValidator, VistockConverter
from typing import List, Dict, Any
from datetime import datetime, timezone
import asyncio

class VistockDNSEStockIndexSearch(IVistockDNSEStockIndexSearch, AsyncIVistockDNSEStockIndexSearch):
    def __init__(self, timeout: float = DEFAULT_TIMEOUT, **kwargs: Any) -> None:
        if timeout <= 0:
            raise ValueError(
                'Invalid configuration: "timeout" must be a strictly positive integer value representing the maximum allowable wait time for the operation.'
            )
        self._timeout = timeout

        if 'semaphore_limit' in kwargs and (not isinstance(kwargs['semaphore_limit'], int) or kwargs['semaphore_limit'] <= 0):
            raise ValueError(
                'Invalid configuration: "semaphore_limit" must be a positive integer, indicating the maximum number of concurrent asynchronous operations permitted.'
            )

        self._semaphore_limit = kwargs.get('semaphore_limit', 5)
        self._base_url = DEFAULT_DNSE_STOCK_INDEX_BASE_URL
        self._domain = DEFAULT_DNSE_DOMAIN
        self._scraper = VistockDNSEStockIndexScraper()
        self._parser = VistockDNSEStockIndexParser()
        self._semaphore = asyncio.Semaphore(self._semaphore_limit)

    @property
    def timeout(self) -> float:
        return self._timeout
    
    @timeout.setter
    def timeout(self, value: int) -> None:
        if value <= 0:
            raise ValueError(
                'Invalid value: "timeout" must be a positive integer greater than zero.'
            )
        self._timeout = value

    def search(
        self,
        code: str,
        current_datetime: str = datetime.now(timezone.utc).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z',
        ascending: bool = False
    ) -> StandardDNSEStockIndexSearchResults:
            payload = self._parser.parse_payload(
                code=code,
                current_datetime=current_datetime
            )

            data: List[Dict[str, Any]] = self._scraper.fetch(url=self._base_url, payload=payload).get('data', {}).get('GetKrxTicksBySymbols', {}).get('ticks', [])

            if not data:
                raise ValueError(
                    'No data found for the given parameters. Please check the code, and current date to ensure they are correct and that data exists for the specified range.'
                )
            
            if not VistockValidator.validate_dnse_stock_index_json_data(data):
                raise ValueError(
                    'Invalid data format: The fetched data does not conform to the expected JSON structure. Please ensure that the API response is valid and contains the necessary fields.'
                )
            
            data.sort(key=lambda x: x.get('sendingTime', ''), reverse=not ascending)

            return StandardDNSEStockIndexSearchResults(
                results=[
                    StandardDNSEStockIndexSearch(
                        code=item.get('symbol', ''),
                        match_price=item.get('matchPrice', 0.0),
                        match_volume= item.get('matchQtty', 0),
                        sending_time=VistockConverter.convert_utc_to_local(item.get('sendingTime', '')),
                        side=item.get('side', 0)
                    ) for item in data
                ],
                total_results=len(data)
            )
    
    async def async_search(
        self,
        code: str,
        current_datetime: str = datetime.now(timezone.utc).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z',
        ascending: bool = False
    ) -> StandardDNSEStockIndexSearchResults:
        async with self._semaphore:
            payload = self._parser.parse_payload(
                code=code,
                current_datetime=current_datetime
            )

            response = await self._scraper.async_fetch(url=self._base_url, payload=payload)
            data: List[Dict[str, Any]] = response.get('data', {}).get('GetKrxTicksBySymbols', {}).get('ticks', [])

            if not data:
                raise ValueError(
                    'No data found for the given parameters. Please check the code, start date, and end date to ensure they are correct and that data exists for the specified range.'
                )
            
            if not VistockValidator.validate_dnse_stock_index_json_data(data):
                raise ValueError(
                    'Invalid data format: The fetched data does not conform to the expected JSON structure. Please ensure that the API response is valid and contains the necessary fields.'
                )
            
            data.sort(key=lambda x: x.get('sendingTime', ''), reverse=not ascending)

            return StandardDNSEStockIndexSearchResults(
                results=[
                    StandardDNSEStockIndexSearch(
                        code=item.get('symbol', ''),
                        match_price=item.get('matchPrice', 0.0),
                        match_volume= item.get('matchQtty', 0),
                        sending_time=VistockConverter.convert_utc_to_local(item.get('sendingTime', '')),
                        side=item.get('side', 0)
                    ) for item in data
                ],
                total_results=len(data)
            )
            

            




class InvalidURLError(Exception):
    def __init__(self, message: str = "The provided URL is invalid. Please ensure it is properly formatted."):
        self.message = message
        super().__init__(self.message)


from vistock.core.interfaces.ivistockparser import IVistockVietstockStockIndexParser
from vistock.core.constants import DEFAULT_VIETSTOCK_DOMAIN
from vistock.core.utils import VistockValidator, VistockConverter
from datetime import datetime

class VistockVietstockStockIndexParser(IVistockVietstockStockIndexParser):
    def __init__(self):
        self._domain = DEFAULT_VIETSTOCK_DOMAIN

    def parse_url_path(
        self,
        code: str,
        resolution: str,
        start_date: str = '2000-01-01',
        end_date: str = datetime.now().strftime('%Y-%m-%d')
    ) -> str:
        if not VistockValidator.validate_code(code=code):
            raise ValueError(
                'Invalid code: "code" must be a non-empty alphanumeric string with exactly 3 characters representing the stock code. Please ensure that the code is specified correctly.'
            )
        
        if not VistockValidator.validate_vietstock_resolution(resolution=resolution):
            raise ValueError(
                'Invalid resolution: "resolution" must be "1D". Please ensure that the resolution is specified correctly.'
            )
        
        if not VistockValidator.validate_date_range(start_date=start_date, end_date=end_date):
            raise ValueError(
                'Invalid date range: "start_date" must be earlier than "end_date". Please ensure that the start date precedes the end date to maintain a valid chronological order.'
            )
        
        if not VistockValidator.validate_date_format(date_str=start_date):
            raise ValueError(
                f'Invalid start_date format: "{start_date}". Please use "YYYY-MM-DD".'
            )
        start_date_timestamp: int = VistockConverter.convert_date_to_timestamp(date=start_date)

        if not VistockValidator.validate_date_format(date_str=end_date):
            raise ValueError(
                f'Invalid end_date format: "{end_date}". Please use "YYYY-MM-DD".'
            )
        end_date_timestamp: int = VistockConverter.convert_date_to_timestamp(date=end_date)

        return f'?symbol={code}&resolution={resolution}&from={start_date_timestamp}&to={end_date_timestamp}&countback=2'




from vistock.core.constants import (
    DEFAULT_24HMONEY_BASE_URL,
    DEFAULT_24HMONEY_DOMAIN,
    DEFAULT_TIMEOUT
)
from vistock.core.models import (
    Standard24HMoneyStockSection,
    Standard24HMoneyStockSectionSearchResults
)
from vistock.core.enums import (
    Vistock24HMoneyIndustryCategory,
    Vistock24HMoneyFloorCategory,
    Vistock24HMoneyCompanyCategory,
    Vistock24HMoneyLetterCategory
)
from vistock.modules._24hmoney.scrapers import Vistock24HMoneyStockSectionScraper
from vistock.modules._24hmoney.parsers import Vistock24HMoneyStockSectionParser
from vistock.core.interfaces.ivistocksearch import (
    IVistock24HMoneyStockSectionSearch,
    AsyncIVistock24HMoneyStockSectionSearch
)
from vistock.core.utils import VistockValidator, VistockNormalizator, VistockMapper
from typing import List, Dict, Union, Any
import asyncio

class Vistock24HMoneyStockSectionSearch(IVistock24HMoneyStockSectionSearch, AsyncIVistock24HMoneyStockSectionSearch):
    def __init__(self, timeout: float = DEFAULT_TIMEOUT, **kwargs: Any) -> None:
        if timeout <= 0:
            raise ValueError(
                'Invalid configuration: "timeout" must be a strictly positive integer value representing the maximum allowable wait time for the operation.'
            )
        self._timeout = timeout

        if 'semaphore_limit' in kwargs and (not isinstance(kwargs['semaphore_limit'], int) or kwargs['semaphore_limit'] <= 0):
            raise ValueError(
                'Invalid configuration: "semaphore_limit" must be a positive integer, indicating the maximum number of concurrent asynchronous operations permitted.'
            )

        self._semaphore_limit = kwargs.get('semaphore_limit', 5)
        self._base_url = DEFAULT_24HMONEY_BASE_URL
        self._domain = DEFAULT_24HMONEY_DOMAIN
        self._scraper = Vistock24HMoneyStockSectionScraper()
        self._parser = Vistock24HMoneyStockSectionParser()
        self._semaphore = asyncio.Semaphore(self._semaphore_limit)

    @property
    def timeout(self) -> float:
        return self._timeout
    
    @timeout.setter
    def timeout(self, value: int) -> None:
        if value <= 0:
            raise ValueError(
                'Invalid value: "timeout" must be a positive integer greater than zero.'
            )
        self._timeout = value

    def search(
        self,
        industry: Union[Vistock24HMoneyIndustryCategory, str] = 'all',
        floor: Union[Vistock24HMoneyFloorCategory, str] = 'all',
        company_type: Union[Vistock24HMoneyCompanyCategory, str] = 'all',
        letter: Union[Vistock24HMoneyLetterCategory, str] = 'all',
        limit: int = 2000
    ) -> Standard24HMoneyStockSectionSearchResults:
        if industry != 'all':
            if not VistockValidator.validate_enum_value(industry, Vistock24HMoneyIndustryCategory):
                raise ValueError(f'"{industry}" is not a recognized industry. Use a valid enum name or code.')
            industry_code = VistockNormalizator.normalize_enum_value(industry, Vistock24HMoneyIndustryCategory)
        else:
            industry_code = 'all'

        if floor != 'all':
            if not VistockValidator.validate_enum_value(floor, Vistock24HMoneyFloorCategory):
                raise ValueError(f'"{floor}" is not a valid floor. Please use a correct market floor name or code.')
            floor_code = VistockNormalizator.normalize_enum_value(floor, Vistock24HMoneyFloorCategory)
        else:
            floor_code = 'all'

        if company_type != 'all':
            if not VistockValidator.validate_enum_value(company_type, Vistock24HMoneyCompanyCategory):
                raise ValueError(f'"{company_type}" is not a valid company type. Use a defined category or symbol.')
            company_code = VistockNormalizator.normalize_enum_value(company_type, Vistock24HMoneyCompanyCategory)
        else:
            company_code = 'all'

        if letter != 'all':
            if not VistockValidator.validate_enum_value(letter, Vistock24HMoneyLetterCategory):
                raise ValueError(f'"{letter}" is not a valid letter. Only A-Z are accepted.')
            letter_code = VistockNormalizator.normalize_enum_value(letter, Vistock24HMoneyLetterCategory)
        else:
            letter_code = 'all'

        url = f'{self._base_url}{self._parser.parse_url_path(industry_code=industry_code, floor_code=floor_code, company_code=company_code, letter_code=letter_code, limit=limit)}'
        data: List[Dict[str, Any]] = self._scraper.fetch(url=url).get('data', []).get('data', [])

        results: List[Standard24HMoneyStockSection] = []
        for item in data:
            icb_name_vi = item.get('icb_name_vi', '')
            try:
                icb_name_en = VistockMapper.map_english_section(vn_section=icb_name_vi)
            except ValueError:
                icb_name_en = ''

            result = Standard24HMoneyStockSection(
                code=item.get('symbol', ''),
                company_name=item.get('company_name', ''),
                tfloor=item.get('floor', ''),
                company_type=item.get('fiingroup_com_type_code', '') or '',
                icb_name_vi=icb_name_vi or '',
                icb_name_en=icb_name_en,
                listed_share_vol=item.get('listed_share_vol', 0) or 0,
                fiingroup_icb_code=item.get('fiingroup_icb_code', 0)
            )
            results.append(result)
        
        return Standard24HMoneyStockSectionSearchResults(results=results, total_results=len(data))
        
    async def async_search(
        self,
        industry: Union[Vistock24HMoneyIndustryCategory, str] = 'all',
        floor: Union[Vistock24HMoneyFloorCategory, str] = 'all',
        company_type: Union[Vistock24HMoneyCompanyCategory, str] = 'all',
        letter: Union[Vistock24HMoneyLetterCategory, str] = 'all',
        limit: int = 2000
    ) -> Standard24HMoneyStockSectionSearchResults:
        if industry != 'all':
            if not VistockValidator.validate_enum_value(industry, Vistock24HMoneyIndustryCategory):
                raise ValueError(f'"{industry}" is not a recognized industry. Use a valid enum name or code.')
            industry_code = VistockNormalizator.normalize_enum_value(industry, Vistock24HMoneyIndustryCategory)
        else:
            industry_code = 'all'

        if floor != 'all':
            if not VistockValidator.validate_enum_value(floor, Vistock24HMoneyFloorCategory):
                raise ValueError(f'"{floor}" is not a valid floor. Please use a correct market floor name or code.')
            floor_code = VistockNormalizator.normalize_enum_value(floor, Vistock24HMoneyFloorCategory)
        else:
            floor_code = 'all'

        if company_type != 'all':
            if not VistockValidator.validate_enum_value(company_type, Vistock24HMoneyCompanyCategory):
                raise ValueError(f'"{company_type}" is not a valid company type. Use a defined category or symbol.')
            company_code = VistockNormalizator.normalize_enum_value(company_type, Vistock24HMoneyCompanyCategory)
        else:
            company_code = 'all'

        if letter != 'all':
            if not VistockValidator.validate_enum_value(letter, Vistock24HMoneyLetterCategory):
                raise ValueError(f'"{letter}" is not a valid letter. Only A-Z are accepted.')
            letter_code = VistockNormalizator.normalize_enum_value(letter, Vistock24HMoneyLetterCategory)
        else:
            letter_code = 'all'

        url = f'{self._base_url}{self._parser.parse_url_path(industry_code=industry_code, floor_code=floor_code, company_code=company_code, letter_code=letter_code, limit=limit)}'

        async with self._semaphore:
            raw_response = await self._scraper.async_fetch(url=url)  # This line assumes your scraper is async

        data: List[Dict[str, Any]] = raw_response.get('data', []).get('data', [])

        results: List[Standard24HMoneyStockSection] = []
        for item in data:
            icb_name_vi = item.get('icb_name_vi', '')
            try:
                icb_name_en = VistockMapper.map_english_section(vn_section=icb_name_vi)
            except ValueError:
                icb_name_en = ''

            result = Standard24HMoneyStockSection(
                code=item.get('symbol', ''),
                company_name=item.get('company_name', ''),
                tfloor=item.get('floor', ''),
                company_type=item.get('fiingroup_com_type_code', ''),
                icb_name_vi=icb_name_vi,
                icb_name_en=icb_name_en,
                listed_share_vol=item.get('listed_share_vol', 0),
                fiingroup_icb_code=item.get('fiingroup_icb_code', 0)
            )
            results.append(result)

        return Standard24HMoneyStockSectionSearchResults(results=results, total_results=len(data))


from vistock.core.constants import (
    DEFAULT_VIETSTOCK_STOCK_INDEX_BASE_URL,
    DEFAULT_VIETSTOCK_DOMAIN,
    DEFAULT_TIMEOUT
)
from vistock.core.models import (
    StandardVietstockStockIndexSearch,
    StandardVietstockStockIndexSearchResults
)
from vistock.core.interfaces.ivistocksearch import (
    IVistockVietstockStockIndexSearch,
    AsyncIVistockVietstockStockIndexSearch
)
from vistock.modules.vietstock.scrapers import VistockVietstockStockIndexScraper
from vistock.modules.vietstock.parsers import VistockVietstockStockIndexParser
from vistock.core.utils import VistockConverter
from typing import List, Dict, Literal, Any
from datetime import datetime
import asyncio

class VistockVietstockStockIndexSearch(IVistockVietstockStockIndexSearch, AsyncIVistockVietstockStockIndexSearch):
    def __init__(self, timeout: float = DEFAULT_TIMEOUT, **kwargs: Any) -> None:
        if timeout <= 0:
            raise ValueError(
                'Invalid configuration: "timeout" must be a strictly positive integer value representing the maximum allowable wait time for the operation.'
            )
        self._timeout = timeout

        if 'semaphore_limit' in kwargs and (not isinstance(kwargs['semaphore_limit'], int) or kwargs['semaphore_limit'] <= 0):
            raise ValueError(
                'Invalid configuration: "semaphore_limit" must be a positive integer, indicating the maximum number of concurrent asynchronous operations permitted.'
            )

        self._semaphore_limit = kwargs.get('semaphore_limit', 5)
        self._base_url = DEFAULT_VIETSTOCK_STOCK_INDEX_BASE_URL
        self._domain = DEFAULT_VIETSTOCK_DOMAIN
        self._scraper = VistockVietstockStockIndexScraper()
        self._parser = VistockVietstockStockIndexParser()
        self._semaphore = asyncio.Semaphore(self._semaphore_limit)

    @property
    def timeout(self) -> float:
        return self._timeout
    
    @timeout.setter
    def timeout(self, value: int) -> None:
        if value <= 0:
            raise ValueError(
                'Invalid value: "timeout" must be a positive integer greater than zero.'
            )
        self._timeout = value

    def search(
        self,
        code: str,
        resolution: Literal['1D'] = '1D',
        start_date: str = '2000-01-01',
        end_date: str = datetime.now().strftime('%Y-%m-%d'),
        ascending: bool = False
    ) -> StandardVietstockStockIndexSearchResults:
        url = f'{self._base_url}{self._parser.parse_url_path(code=code, resolution=resolution, start_date=start_date, end_date=end_date)}'
        response: Dict[str, Any] = self._scraper.fetch(url=url)

        data: List[StandardVietstockStockIndexSearch] = []
        if response.get('s') == 'ok':
            mopens = response.get('o', [])
            mhighs = response.get('h', [])
            mlows = response.get('l', [])
            mcloses = response.get('c', [])
            mvolumes = response.get('v', [])
            timestamps = response.get('t', [])

            for mopen, mhigh, mlow, mclose, mvolume, timestamp in zip(mopens, mhighs, mlows, mcloses, mvolumes, timestamps):
                data.append(StandardVietstockStockIndexSearch(
                    mopen=mopen,
                    mhigh=mhigh,
                    mlow=mlow,
                    mclose=mclose,
                    mvolume=int(mvolume),
                    timestamp=VistockConverter.convert_timestamp_to_date(timestamp),
                ))

        data.sort(key=lambda x: x.timestamp, reverse=not ascending)

        return StandardVietstockStockIndexSearchResults(
            results=data,
            total_results=len(data)
        )
    
    async def async_search(
        self,
        code: str,
        resolution: Literal['1D'] = '1D',
        start_date: str = '2000-01-01',
        end_date: str = datetime.now().strftime('%Y-%m-%d'),
        ascending: bool = False
    ) -> StandardVietstockStockIndexSearchResults:
        url = f'{self._base_url}{self._parser.parse_url_path(code=code, resolution=resolution, start_date=start_date, end_date=end_date)}'
        response: Dict[str, Any] = await self._scraper.async_fetch(url=url)

        data: List[StandardVietstockStockIndexSearch] = []
        if response.get('s') == 'ok':
            mopens = response.get('o', [])
            mhighs = response.get('h', [])
            mlows = response.get('l', [])
            mcloses = response.get('c', [])
            mvolumes = response.get('v', [])
            timestamps = response.get('t', [])

            for mopen, mhigh, mlow, mclose, mvolume, timestamp in zip(mopens, mhighs, mlows, mcloses, mvolumes, timestamps):
                data.append(StandardVietstockStockIndexSearch(
                    mopen=mopen,
                    mhigh=mhigh,
                    mlow=mlow,
                    mclose=mclose,
                    mvolume=int(mvolume),
                    timestamp=VistockConverter.convert_timestamp_to_date(timestamp),
                ))

        data.sort(key=lambda x: x.timestamp, reverse=not ascending)

        return StandardVietstockStockIndexSearchResults(
            results=data,
            total_results=len(data)
        ) 




from .core.enums import (
    Vistock24HMoneyIndustryCategory,
    Vistock24HMoneyFloorCategory,
    Vistock24HMoneyCompanyCategory,
    Vistock24HMoneyLetterCategory,
    VistockVnDirectFinancialModelsCategory,
    VistockVnDirectReportTypeCategory,
    VistockVnDirectIndexCodeMapping,
    VistockVnDirectChangePricePeriodMapping
)

__all__ = [
    'Vistock24HMoneyIndustryCategory',
    'Vistock24HMoneyFloorCategory',
    'Vistock24HMoneyCompanyCategory',
    'Vistock24HMoneyLetterCategory',
    'VistockVnDirectFinancialModelsCategory',
    'VistockVnDirectReportTypeCategory',
    'VistockVnDirectIndexCodeMapping',
    'VistockVnDirectChangePricePeriodMapping'
]


from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='vistock',
    version='0.1.0-beta.4',
    install_requires=requirements,
    include_package_data=True,
    packages=find_packages(exclude=['tests', 'tests.*']),
    author='Thang Duong',
    author_email="kaiismith.business@gmail.com",
    description='Vistock is an open-source library focused on searching, retrieving, and analyzing Vietnamese stock market data.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache-2.0',
    keywords='stock, vietnamese, vietnam, financial data, stock market',
    python_requires='>=3.12',
    url='https://github.com/kaiismith/vistock',
    project_urls={
        'Bug Reports': 'https://github.com/kaiismith/vistock/issues',
        'Source': 'https://github.com/kaiismith/vistock',
        'Documentation': 'https://github.com/kaiismith/vistock#readme',
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Programming Language :: Python :: 3.15',
        'Topic :: Software Development :: Libraries',
        'Topic :: Office/Business :: Financial',
    ],
)


from vistock.core.interfaces.ivistockparser import IVistock24HMoneyStockSectionParser
from vistock.core.constants import DEFAULT_24HMONEY_DOMAIN
from vistock.core.enums import (
    Vistock24HMoneyIndustryCategory,
    Vistock24HMoneyFloorCategory,
    Vistock24HMoneyCompanyCategory,
    Vistock24HMoneyLetterCategory
)
from typing import Union

class Vistock24HMoneyStockSectionParser(IVistock24HMoneyStockSectionParser):
    def __init__(self):
        self._domain = DEFAULT_24HMONEY_DOMAIN

    def parse_url_path(
        self,
        industry_code: Union[Vistock24HMoneyIndustryCategory, str] = 'all',
        floor_code: Union[Vistock24HMoneyFloorCategory, str] = 'all',
        company_code: Union[Vistock24HMoneyCompanyCategory, str] = 'all',
        letter_code: Union[Vistock24HMoneyLetterCategory, str] = 'all',
        limit: int = 2000
    ) -> str:
        if limit < 0:
            raise ValueError(
                'Invalid limit: "limit" must be a positive integer greater than zero to ensure proper pagination and data retrieval.'
            )
        
        if limit == 0:
            limit += 1

        return f'?&industry_code={industry_code}&floor_code={floor_code}&com_type={company_code}&letter={letter_code}&page=1&per_page={limit}'


from vistock.core.interfaces.ivistockparser import IVistockDNSEStockIndexParser
from vistock.core.constants import DEFAULT_DNSE_DOMAIN
from vistock.core.utils import VistockValidator
from datetime import datetime, timezone, timedelta
from typing import Dict, Any

class VistockDNSEStockIndexParser(IVistockDNSEStockIndexParser):
    def __init__(self):
        self._domain = DEFAULT_DNSE_DOMAIN

    def parse_payload(
        self,
        code: str,
        current_datetime: str = datetime.now(timezone.utc).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'
    ) -> Dict[str, Any]:
        if not VistockValidator.validate_code(code=code):
            raise ValueError(
                'Invalid code: "code" must be a non-empty alphanumeric string with exactly 3 characters representing the stock code. Please ensure that the code is specified correctly.'
            )
        
        dt = datetime.strptime(current_datetime, '%Y-%m-%dT%H:%M:%S.%fZ')

        if dt.weekday() == 5:
            dt -= timedelta(days=1)
        elif dt.weekday() == 6:
            dt -= timedelta(days=2)

        date = dt.strftime('%Y-%m-%d')
        current_datetime = dt.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'

        return {
            'query': f'''
                query GetKrxTicksBySymbols {{
                    GetKrxTicksBySymbols(
                        symbols: "{code}", 
                        date: "{date}", 
                        limit: 100000, 
                        before: "{current_datetime}", 
                        board: 2
                    ) {{
                        ticks {{
                            symbol
                            matchPrice
                            matchQtty
                            sendingTime
                            side
                        }}
                    }}
                }}
            '''
        }
        


from pydantic import BaseModel
from typing import List, Optional

class StandardVnDirectStockIndexSearch(BaseModel):
    code: str
    date: str
    time: str
    tfloor: str
    type: str
    mopen: float
    mhigh: float
    mlow: float
    mclose: float
    maverage: float
    nmvolume: int

    def __repr__(self):
        return super().__repr__()

class AdvancedVnDirectStockIndexSearch(BaseModel):
    standard: StandardVnDirectStockIndexSearch
    basic: float
    ceiling: float
    floor: float
    open: float
    high: float
    low: float
    close: float
    average: float
    nmvalue: float
    ptvolume: float
    ptvalue: float
    change: float
    mchange: float
    pctchange: float

    def __repr__(self):
        return super().__repr__()

class StandardVnDirectStockIndexSearchResults(BaseModel):
    results: List[StandardVnDirectStockIndexSearch]
    total_results: int

    def __str__(self):
        results_repr = ', '.join(repr(r) for r in self.results)
        return f'{self.__class__.__name__}(results=[{results_repr}], total_results={self.total_results})'

class AdvancedVnDirectStockIndexSearchResults(BaseModel):
    results: List[AdvancedVnDirectStockIndexSearch]
    total_results: int

    def __str__(self):
        results_repr = ', '.join(repr(r) for r in self.results)
        return f'{self.__class__.__name__}(results=[{results_repr}], total_results={self.total_results})'

class StandardVnDirectFundamentalIndexSearchResults(BaseModel):
    marketcap: float
    nm_volume_avg_cr_10d: float
    price_highest_cr_52w: float
    price_lowest_cr_52w: float
    outstanding_shares: float
    freefloat: float
    beta: float
    price_to_earnings: float
    price_to_book: float
    roae_tr_avg_5q: float
    roaa_tr_avg_5q: float
    dividend_yield: float
    eps_tr: float
    bvps_cr: float

    def __str__(self):
        fields = ', '.join(f"{k}={v!r}" for k, v in self.model_dump().items())
        return f'{self.__class__.__name__}({fields})'
    
class StandardVnDirectFinancialModelSearch(BaseModel):
    model_type: int
    model_type_name: str
    model_vn_desc: str
    model_en_desc: str
    company_form: str
    note: str
    item_code: int
    item_vn_name: str
    item_en_name: str
    display_order: int
    display_level: int
    form_type: str 

    def __repr__(self):
        return super().__repr__()

class StandardVnDirectFinancialModelSearchResults(BaseModel):
    results: List[StandardVnDirectFinancialModelSearch]
    total_results: int

    def __str__(self):
        results_repr = ', '.join(repr(r) for r in self.results)
        return f'{self.__class__.__name__}(results=[{results_repr}], total_results={self.total_results})'
    
class StandardVnDirectFinancialStatementsIndex(BaseModel):
    code: str
    model: StandardVnDirectFinancialModelSearch
    report_type: str
    numeric_value: int
    fiscal_date: str
    created_date: str
    modified_date: str

    def __repr__(self):
        return super().__repr__()

class StandardVnDirectFinancialStatementsIndexSearchResults(BaseModel):
    results: List[StandardVnDirectFinancialStatementsIndex]
    total_results: int

    def __str__(self):
        results_repr = ', '.join(repr(r) for r in self.results)
        return f'{self.__class__.__name__}(results=[{results_repr}], total_results={self.total_results})'
    
class StandardVnDirectMarketPricesSearch(BaseModel):
    code: str
    date: str
    time: str
    tfloor: str
    type: str
    open: float
    high: float
    low: float
    close: float
    change: float
    pct_change: float
    accumulated_volume: float
    accumulated_value: float
    nmvolume: float
    nmvalue: float
    ptvolume: float
    ptvalue: float
    advances: float
    declines: float
    no_change: float
    no_trade: float
    ceiling_stocks: float
    floor_stocks: float
    val_chg_pct_cr1d: float
    
    def __repr__(self):
        return super().__repr__()

class StandardVnDirectMarketPricesSearchResults(BaseModel):
    results: List[StandardVnDirectMarketPricesSearch]
    total_results: int

    def __str__(self):
        results_repr = ', '.join(repr(r) for r in self.results)
        return f'{self.__class__.__name__}(results=[{results_repr}], total_results={self.total_results})'
        
class StandardVnDirectChangePricesSearch(BaseModel):
    code: str
    name: str
    type: str
    period: str
    price: float
    bop_price: float
    change: float
    pct_change: float
    last_updated: str

    def __repr__(self):
        return super().__repr__()
    
class StandardVnDirectChangePricesSearchResults(BaseModel):
    results: List[StandardVnDirectChangePricesSearch]
    total_results: int

    def __str__(self):
        results_repr = ', '.join(repr(r) for r in self.results)
        return f'{self.__class__.__name__}(results=[{results_repr}], total_results={self.total_results})'
    
class Standard24HMoneyStockSection(BaseModel):
    code: str
    company_name: str
    tfloor: str
    company_type: Optional[str] = ''
    icb_name_vi: Optional[str] = ''
    icb_name_en: Optional[str] = ''
    listed_share_vol: Optional[int] = 0
    fiingroup_icb_code: int

    def __repr__(self):
        return super().__repr__()

class Standard24HMoneyStockSectionSearchResults(BaseModel):
    results: List[Standard24HMoneyStockSection]
    total_results: int

    def __str__(self):
        results_repr = ', '.join(repr(r) for r in self.results)
        return f'{self.__class__.__name__}(results=[{results_repr}], total_results={self.total_results})'
    
class StandardVietstockStockIndexSearch(BaseModel):
    mopen: float
    mhigh: float
    mlow: float
    mclose: float
    mvolume: int
    timestamp: str

    def __repr__(self):
        return super().__repr__()

class StandardVietstockStockIndexSearchResults(BaseModel):
    results: List[StandardVietstockStockIndexSearch]
    total_results: int

    def __str__(self):
        results_repr = ', '.join(repr(r) for r in self.results)
        return f'{self.__class__.__name__}(results=[{results_repr}], total_results={self.total_results})'
    
class StandardDNSEStockIndexSearch(BaseModel):
    code: str
    match_price: float
    match_volume: int
    sending_time: str
    side: int

    def __repr__(self):
        return super().__repr__()

class StandardDNSEStockIndexSearchResults(BaseModel):
    results: List[StandardDNSEStockIndexSearch]
    total_results: int

    def __str__(self):
        results_repr = ', '.join(repr(r) for r in self.results)
        return f'{self.__class__.__name__}(results=[{results_repr}], total_results={self.total_results})'
